#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, List

from fabric_mb.message_bus.messages.poa_avro import PoaAvro
from fabric_mb.message_bus.messages.proxy_avro import ProxyAvro
from fabric_mb.message_bus.messages.reservation_predecessor_avro import ReservationPredecessorAvro
from fabric_mb.message_bus.messages.result_delegation_avro import ResultDelegationAvro
from fim.slivers.base_sliver import BaseSliver
from fim.user import GraphFormat

from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.manage.client_actor_management_object_helper import ClientActorManagementObjectHelper
from fabric_cf.actor.core.manage.proxy_protocol_descriptor import ProxyProtocolDescriptor
from fabric_cf.actor.core.manage.server_actor_management_object import ServerActorManagementObject
from fabric_cf.actor.core.apis.abc_client_actor_management_object import ABCClientActorManagementObject

if TYPE_CHECKING:
    from fabric_mb.message_bus.messages.result_broker_query_model_avro import ResultBrokerQueryModelAvro
    from fabric_mb.message_bus.messages.result_string_avro import ResultStringAvro
    from fabric_mb.message_bus.messages.ticket_reservation_avro import TicketReservationAvro
    from fabric_mb.message_bus.messages.result_strings_avro import ResultStringsAvro
    from fabric_mb.message_bus.messages.reservation_mng import ReservationMng
    from fabric_mb.message_bus.messages.result_avro import ResultAvro
    from fabric_mb.message_bus.messages.result_proxy_avro import ResultProxyAvro

    from fabric_cf.actor.core.apis.abc_broker_mixin import ABCBrokerMixin
    from fabric_cf.actor.core.apis.abc_actor_mixin import ABCActorMixin
    from fabric_cf.actor.security.auth_token import AuthToken
    from fabric_cf.actor.core.util.id import ID


class BrokerManagementObject(ServerActorManagementObject, ABCClientActorManagementObject):
    def __init__(self, *, broker: ABCBrokerMixin = None):
        super().__init__(sa=broker)
        self.client_helper = ClientActorManagementObjectHelper(client=broker)

    def register_protocols(self):
        from fabric_cf.actor.core.manage.local.local_broker import LocalBroker
        local = ProxyProtocolDescriptor(protocol=Constants.PROTOCOL_LOCAL,
                                        proxy_class=LocalBroker.__name__,
                                        proxy_module=LocalBroker.__module__)

        from fabric_cf.actor.core.manage.kafka.kafka_broker import KafkaBroker
        kafka = ProxyProtocolDescriptor(protocol=Constants.PROTOCOL_KAFKA,
                                        proxy_class=KafkaBroker.__name__,
                                        proxy_module=KafkaBroker.__module__)

        self.proxies = []
        self.proxies.append(local)
        self.proxies.append(kafka)

    def save(self) -> dict:
        properties = super().save()
        properties[Constants.PROPERTY_CLASS_NAME] = BrokerManagementObject.__name__
        properties[Constants.PROPERTY_MODULE_NAME] = BrokerManagementObject.__module__

        return properties

    def set_actor(self, *, actor: ABCActorMixin):
        if self.actor is None:
            super().set_actor(actor=actor)
            self.client_helper = ClientActorManagementObjectHelper(client=actor)

    def get_brokers(self, *, caller: AuthToken, broker_id: ID = None) -> ResultProxyAvro:
        return self.client_helper.get_brokers(caller=caller, broker_id=broker_id)

    def add_broker(self, *, broker: ProxyAvro, caller: AuthToken) -> ResultAvro:
        return self.client_helper.add_broker(broker=broker, caller=caller)

    def update_broker(self, *, broker: ProxyAvro, caller: AuthToken) -> ResultAvro:
        return self.client_helper.update_broker(broker=broker, caller=caller)

    def get_broker_query_model(self, *, broker: ID, caller: AuthToken, id_token: str,
                               level: int, graph_format: GraphFormat, start: datetime = None,
                               end: datetime = None, includes: str = None,
                               excludes: str = None) -> ResultBrokerQueryModelAvro:
        return self.client_helper.get_broker_query_model(broker=broker, caller=caller, id_token=id_token, level=level,
                                                         graph_format=graph_format, start=start, end=end,
                                                         includes=includes, excludes=excludes)

    def add_reservation(self, *, reservation: TicketReservationAvro, caller: AuthToken) -> ResultStringAvro:
        return self.client_helper.add_reservation(reservation=reservation, caller=caller)

    def add_reservations(self, *, reservations: List[TicketReservationAvro], caller: AuthToken) -> ResultStringsAvro:
        return self.client_helper.add_reservations(reservations=reservations, caller=caller)

    def demand_reservation_rid(self, *, rid: ID, caller: AuthToken) -> ResultAvro:
        return self.client_helper.demand_reservation_rid(rid=rid, caller=caller)

    def demand_reservation(self, *, reservation: ReservationMng, caller: AuthToken) -> ResultAvro:
        return self.client_helper.demand_reservation(reservation=reservation, caller=caller)

    def extend_reservation(self, *, reservation: ID, new_end_time: datetime, sliver: BaseSliver,
                           caller: AuthToken, dependencies: List[ReservationPredecessorAvro] = None) -> ResultAvro:
        return self.client_helper.extend_reservation(reservation=reservation, new_end_time=new_end_time,
                                                     sliver=sliver, caller=caller, dependencies=dependencies)

    def claim_delegations(self, *, broker: ID, did: str, caller: AuthToken) -> ResultDelegationAvro:
        return self.client_helper.claim_delegations(broker=broker, did=did, caller=caller)

    def reclaim_delegations(self, *, broker: ID, did: str, caller: AuthToken) -> ResultDelegationAvro:
        return self.client_helper.reclaim_delegations(broker=broker, did=did, caller=caller)

    def modify_reservation(self, *, rid: ID, modified_sliver: BaseSliver, caller: AuthToken) -> ResultAvro:
        return self.client_helper.modify_reservation(rid=rid, modified_sliver=modified_sliver, caller=caller)

    def poa(self, *, poa: PoaAvro, caller: AuthToken) -> ResultAvro:
        return self.client_helper.poa(poa=poa, caller=caller)



# coding: utf-8

from __future__ import absolute_import
from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from fabric_cf.orchestrator.swagger_server.models.base_model_ import Model
from fabric_cf.orchestrator.swagger_server import util


class VersionData(Model):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    def __init__(self, reference: str=None, version: str=None):  # noqa: E501
        """VersionData - a model defined in Swagger

        :param reference: The reference of this VersionData.  # noqa: E501
        :type reference: str
        :param version: The version of this VersionData.  # noqa: E501
        :type version: str
        """
        self.swagger_types = {
            'reference': str,
            'version': str
        }

        self.attribute_map = {
            'reference': 'reference',
            'version': 'version'
        }
        self._reference = reference
        self._version = version

    @classmethod
    def from_dict(cls, dikt) -> 'VersionData':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The version_data of this VersionData.  # noqa: E501
        :rtype: VersionData
        """
        return util.deserialize_model(dikt, cls)

    @property
    def reference(self) -> str:
        """Gets the reference of this VersionData.


        :return: The reference of this VersionData.
        :rtype: str
        """
        return self._reference

    @reference.setter
    def reference(self, reference: str):
        """Sets the reference of this VersionData.


        :param reference: The reference of this VersionData.
        :type reference: str
        """
        if reference is None:
            raise ValueError("Invalid value for `reference`, must not be `None`")  # noqa: E501

        self._reference = reference

    @property
    def version(self) -> str:
        """Gets the version of this VersionData.


        :return: The version of this VersionData.
        :rtype: str
        """
        return self._version

    @version.setter
    def version(self, version: str):
        """Sets the version of this VersionData.


        :param version: The version of this VersionData.
        :type version: str
        """
        if version is None:
            raise ValueError("Invalid value for `version`, must not be `None`")  # noqa: E501

        self._version = version



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import logging
import pickle
import threading
from contextlib import contextmanager
from datetime import datetime, timezone
from typing import List, Tuple, Dict, Optional

from sqlalchemy import create_engine, desc, func, and_, or_
from sqlalchemy.orm import scoped_session, sessionmaker, joinedload

from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.common.exceptions import DatabaseException
from fabric_cf.actor.db import Base, Clients, ConfigMappings, Proxies, Units, Reservations, Slices, ManagerObjects, \
    Miscellaneous, Actors, Delegations, Sites, Poas, Components, Metrics


@contextmanager
def session_scope(psql_db_engine):
    """Provide a transactional scope around a series of operations."""
    session = scoped_session(sessionmaker(psql_db_engine))
    try:
        yield session
        session.commit()
    except:
        session.rollback()
        raise
    finally:
        session.close()


class PsqlDatabase:
    """
    Implements interface to Postgres database
    """
    OBJECT_NOT_FOUND = "{} Not Found {}"

    def __init__(self, *, user: str, password: str, database: str, db_host: str, logger):
        # Connecting to PostgreSQL server at localhost using psycopg2 DBAPI
        self.db_engine = create_engine("postgresql+psycopg2://{}:{}@{}/{}".format(user, password, db_host, database))
        self.logger = logger
        self.session_factory = sessionmaker(bind=self.db_engine)
        self.sessions = {}

    def get_session(self):
        thread_id = threading.get_ident()
        session = None
        if thread_id in self.sessions:
            session = self.sessions.get(thread_id)
        else:
            session = scoped_session(self.session_factory)
            self.sessions[thread_id] = session
        return session

    def create_db(self):
        """
        Create the database
        """
        Base.metadata.create_all(self.db_engine)

    def set_logger(self, logger):
        """
        Set the logger
        """
        self.logger = logger

    def reset_db(self):
        """
        Reset the database
        """
        session = self.get_session()
        try:
            session.query(Clients).delete()
            session.query(ConfigMappings).delete()
            session.query(Proxies).delete()
            session.query(Units).delete()
            session.query(Delegations).delete()
            session.query(Components).delete()
            session.query(Reservations).delete()
            session.query(Slices).delete()
            session.query(ManagerObjects).delete()
            session.query(Miscellaneous).delete()
            session.query(Actors).delete()
            session.query(Sites).delete()
            session.query(Poas).delete()
            session.query(Metrics).delete()
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def add_actor(self, *, name: str, guid: str, act_type: int, properties):
        """
        Add an actor
        @param name name
        @param guid guid
        @param act_type actor type
        @param properties pickle dump for actor instance
        """
        session = self.get_session()
        try:
            # Save the actor in the database
            actor_obj = Actors(act_name=name, act_guid=guid, act_type=act_type, properties=properties)
            session.add(actor_obj)
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def update_actor(self, *, name: str, properties):
        """
        Update an actor
        @param name name
        @param properties pickle dump for actor instance
        """
        session = self.get_session()
        try:
            actor = session.query(Actors).filter_by(act_name=name).one_or_none()
            if actor is not None:
                actor.properties = properties
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def remove_actor(self, *, name: str):
        """
        Remove an actor
        @param name name
        """
        session = self.get_session()
        try:
            # Delete the actor in the database
            session.query(Actors).filter_by(act_name=name).delete()
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def get_actors(self) -> list:
        """
        Get all actors
        @return list of actors
        """
        result = []
        session = self.get_session()
        try:
            for row in session.query(Actors).all():
                result.append(self.generate_dict_from_row(row=row))
        except Exception as e:
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e
        return result

    def get_actors_by_name_and_type(self, *, actor_name: str, act_type: int) -> list:
        """
        Get actors by name and  actor type
        @param actor_name actor name
        @param act_type actor type
        @return list of actors
        """
        result = []
        session = self.get_session()
        try:
            for row in session.query(Actors).filter(Actors.act_type == act_type).filter(
                    Actors.act_name.like(actor_name)).all():
                result.append(self.generate_dict_from_row(row=row))
        except Exception as e:
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e
        return result

    def get_actors_by_name(self, *, act_name: str) -> list:
        """
        Get actors by name
        @param act_name actor name
        @return list of actors
        """
        result = []
        session = self.get_session()
        try:
            for row in session.query(Actors).filter(Actors.act_name.like(act_name)).all():
                result.append(self.generate_dict_from_row(row=row))
        except Exception as e:
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e
        return result

    def get_actor(self, *, name: str) -> dict:
        """
        Get actor by name
        @return actor
        """
        result = {}
        session = self.get_session()
        try:
            actor = session.query(Actors).filter_by(act_name=name).one_or_none()
            if actor is not None:
                result['act_guid'] = actor.act_guid
                result['act_name'] = actor.act_name
                result['act_type'] = actor.act_type
                result['properties'] = actor.properties
                result['act_id'] = actor.act_id
            else:
                result = None
                self.logger.error("Actor: {} not found!".format(name))
        except Exception as e:
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e
        return result

    def add_miscellaneous(self, *, name: str, properties: dict):
        """
        Add Miscellaneous entries
        @param name name
        @param properties properties

        """
        session = self.get_session()
        try:
            msc_obj = Miscellaneous(msc_path=name, properties=properties)
            session.add(msc_obj)
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def update_miscellaneous(self, *, name: str, properties: dict):
        """
        Add Miscellaneous entries
        @param name name
        @param properties properties

        """
        session = self.get_session()
        try:
            msc_obj = session.query(Miscellaneous).filter_by(msc_path=name).one_or_none()
            if msc_obj is not None:
                msc_obj.properties = properties
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def remove_miscellaneous(self, *, name: str):
        """
        Remove Miscellaneous entries
        @param name name
        """
        session = self.get_session()
        try:
            session.query(Miscellaneous).filter_by(msc_path=name).delete()
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def get_miscellaneous(self, *, name: str) -> dict or None:
        """
        Get Miscellaneous entry
        @param name name
        @return entry identified by name
        """
        result = {}
        session = self.get_session()
        try:
            msc_obj = session.query(Miscellaneous).filter_by(msc_path=name).one_or_none()
            if msc_obj is not None:
                result = msc_obj.properties
            else:
                return None
        except Exception as e:
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e
        return result

    def add_manager_object(self, *, manager_key: str, properties: dict, act_id: int = None):
        """
        Add mananger object
        @param manager_key management object key
        @param properties properties
        @param act_id actor id
        """
        session = self.get_session()
        try:
            if act_id is not None:
                mng_obj = ManagerObjects(mo_key=manager_key, mo_act_id=act_id, properties=properties)
            else:
                mng_obj = ManagerObjects(mo_key=manager_key, properties=properties)
            session.add(mng_obj)
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def remove_manager_object(self, *, manager_key: str):
        """
        Remove management object
        @param manager_key management object key
        """
        session = self.get_session()
        try:
            session.query(ManagerObjects).filter_by(mo_key=manager_key).delete()
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def remove_manager_object_by_actor(self, *, act_id: int):
        """
        Remove management object by actor id
        @param act_id actor id
        """
        session = self.get_session()
        try:
            session.query(ManagerObjects).filter_by(mo_act_id=act_id).delete()
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def get_manager_objects(self, *, act_id: int = None) -> list:
        """
        Get Management objects
        @param act_id actor id
        @return list of objects
        """
        result = []
        session = self.get_session()
        try:
            if act_id is None:
                rows = session.query(ManagerObjects).all()
            else:
                rows = session.query(ManagerObjects).filter_by(mo_act_id=act_id).all()

            for row in rows:
                result.append(self.generate_dict_from_row(row=row))
        except Exception as e:
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e
        return result

    def get_manager_objects_by_actor_name(self, *, act_name: str = None) -> list:
        """
        Get Management objects
        @param act_name actor name
        @return list of objects
        """
        result = []
        try:
            act_obj = self.get_actor(name=act_name)
            act_id = act_obj['act_id']
            result = self.get_manager_objects(act_id=act_id)
        except Exception as e:
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e
        return result

    def get_manager_object(self, *, mo_key: str) -> dict:
        """
        Get Management object by key
        @param mo_key management object key
        @return objects
        """
        result = {}
        session = self.get_session()
        try:
            mo_obj = session.query(ManagerObjects).filter_by(mo_key=mo_key).one_or_none()
            if mo_obj is not None:
                result = self.generate_dict_from_row(mo_obj)
            else:
                raise DatabaseException(self.OBJECT_NOT_FOUND.format("Manager Object", mo_key))
        except Exception as e:
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e
        return result

    def get_manager_containers(self) -> list:
        """
        Get Management object for the container i.e entry with no actor id
        @return object
        """
        result = []
        session = self.get_session()
        try:
            for mo_obj in session.query(ManagerObjects).filter(ManagerObjects.mo_act_id.is_(None)).all():
                result.append(self.generate_dict_from_row(mo_obj))
        except Exception as e:
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e
        return result

    def add_slice(self, *, slc_guid: str, slc_name: str, slc_type: int, slc_state: int, lease_start: datetime = None,
                  lease_end: datetime = None, slc_resource_type: str, properties, slc_graph_id: str = None,
                  oidc_claim_sub: str = None, email: str = None, project_id: str = None):
        """
        Add a slice
        @param slc_guid slice id
        @param slc_name slice name
        @param slc_type slice type
        @param slc_state slice state
        @param slc_resource_type slice resource type
        @param lease_start Lease Start time
        @param lease_end Lease End time
        @param properties pickled instance
        @param slc_graph_id slice graph id
        @param oidc_claim_sub User OIDC Sub
        @param email User Email
        @param project_id Project Id
        """
        session = self.get_session()
        try:
            slc_obj = Slices(slc_guid=slc_guid, slc_name=slc_name, slc_type=slc_type, slc_state=slc_state,
                             oidc_claim_sub=oidc_claim_sub, email=email, slc_resource_type=slc_resource_type,
                             lease_start=lease_start, lease_end=lease_end, properties=properties,
                             project_id=project_id)
            if slc_graph_id is not None:
                slc_obj.slc_graph_id = slc_graph_id
            
            session.add(slc_obj)
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def update_slice(self, *, slc_guid: str, slc_name: str, slc_type: int, slc_state: int, slc_resource_type: str,
                     properties, slc_graph_id: str = None, lease_start: datetime = None, lease_end: datetime = None):
        """
        Update a slice
        @param slc_guid slice id
        @param slc_name slice name
        @param slc_type slice type
        @param slc_state slice state
        @param slc_resource_type slice resource type
        @param lease_start Lease Start time
        @param lease_end Lease End time
        @param properties pickled instance
        @param slc_graph_id slice graph id
        """
        session = self.get_session()
        try:
            slc_obj = session.query(Slices).filter_by(slc_guid=slc_guid).one_or_none()
            if slc_obj is not None:
                slc_obj.properties = properties
                slc_obj.slc_name = slc_name
                slc_obj.slc_type = slc_type
                slc_obj.slc_resource_type = slc_resource_type
                slc_obj.slc_state = slc_state
                slc_obj.lease_start = lease_start
                slc_obj.lease_end = lease_end
                if slc_graph_id is not None:
                    slc_obj.slc_graph_id = slc_graph_id
            else:
                raise DatabaseException(self.OBJECT_NOT_FOUND.format("Slice", slc_guid))
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def remove_slice(self, *, slc_guid: str):
        """
        Remove Slice
        @param slc_guid slice id
        """
        session = self.get_session()
        try:
            session.query(Slices).filter_by(slc_guid=slc_guid).delete()
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def get_slice_ids(self) -> list:
        """
        Get slice ids for an actor
        @param act_id actor id
        @return list of slice ids
        """
        result = []
        session = self.get_session()
        try:
            for row in session.query(Slices).all():
                result.append(row.slc_id)
        except Exception as e:
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e
        return result

    @staticmethod
    def create_slices_filter(*, slice_id: str = None, slice_name: str = None, project_id: str = None, email: str = None,
                             oidc_sub: str = None) -> dict:

        filter_dict = {}
        if slice_id is not None:
            filter_dict['slc_guid'] = slice_id
        if slice_name is not None:
            filter_dict['slc_name'] = slice_name
        if project_id is not None:
            filter_dict['project_id'] = project_id
        if email is not None:
            filter_dict['email'] = email
        if oidc_sub is not None:
            filter_dict['oidc_claim_sub'] = oidc_sub
        return filter_dict

    def get_slice_count(self, *, project_id: str = None, email: str = None, states: List[int] = None,
                        oidc_sub: str = None, slc_type: List[int] = None, excluded_projects: List[str]) -> int:
        """
        Get slices count for an actor
        @param project_id project id
        @param email email
        @param states states
        @param oidc_sub oidc claim sub
        @param slc_type slice type
        @param excluded_projects excluded_projects
        @return list of slices
        """
        session = self.get_session()
        try:
            filter_dict = self.create_slices_filter(project_id=project_id, email=email, oidc_sub=oidc_sub)

            rows = session.query(Slices).filter_by(**filter_dict)

            rows = rows.order_by(desc(Slices.lease_end))

            if states is not None:
                rows = rows.filter(Slices.slc_state.in_(states))

            if slc_type is not None:
                rows = rows.filter(Slices.slc_type.in_(slc_type))

            if excluded_projects is not None:
                rows = rows.filter(Slices.project_id.notin_(excluded_projects))

            return rows.count()
        except Exception as e:
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def get_slices(self, *, slice_id: str = None, slice_name: str = None, project_id: str = None, email: str = None,
                   states: Optional[list[int]] = None, oidc_sub: str = None, slc_type: Optional[list[int]] = None,
                   limit: int = None, offset: int = None, lease_end: datetime = None, search: str = None,
                   exact_match: bool = False, updated_after: datetime = None) -> List[dict]:
        """
        Get slices for an actor, with an option to filter by last update timestamp.

        @param slice_id: actor id
        @param slice_name: slice name
        @param project_id: project id
        @param email: email
        @param states: list of states
        @param oidc_sub: oidc claim sub
        @param slc_type: list of slice types
        @param limit: limit
        @param offset: offset
        @param lease_end: lease_end
        @param search: search term applied
        @param exact_match: Exact Match for Search term
        @param updated_after: Filter slices updated after this timestamp
        @return: list of slices
        """
        result = []
        session = self.get_session()
        try:
            filter_dict = self.create_slices_filter(slice_id=slice_id, slice_name=slice_name,
                                                    project_id=project_id, email=email, oidc_sub=oidc_sub)

            rows = session.query(Slices).filter_by(**filter_dict)

            if search:
                if exact_match:
                    search_term = func.lower(search)
                    rows = rows.filter(((func.lower(Slices.email) == search_term) |
                                        (func.lower(Slices.oidc_claim_sub) == search_term)))
                else:
                    rows = rows.filter(((Slices.email.ilike(f"%{search}%")) |
                                        (Slices.oidc_claim_sub.ilike(f"%{search}%"))))

            if lease_end is not None:
                rows = rows.filter(Slices.lease_end < lease_end)

            if updated_after is not None:
                rows = rows.filter(Slices.last_update_time > updated_after)

            rows = rows.order_by(desc(Slices.lease_end))

            if states is not None:
                rows = rows.filter(Slices.slc_state.in_(states))

            if slc_type is not None:
                rows = rows.filter(Slices.slc_type.in_(slc_type))

            if offset is not None and limit is not None:
                rows = rows.offset(offset).limit(limit)

            for row in rows.all():
                result.append(self.generate_dict_from_row(row=row))
        except Exception as e:
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e
        return result

    def get_slice_by_id(self, *, slc_id: int) -> dict:
        """
        Get slice by id for an actor
        @param slc_id slice id
        @return slice dictionary
        """
        result = {}
        session = self.get_session()
        try:
            slc_obj = session.query(Slices).filter_by(slc_id=slc_id).one_or_none()
            if slc_obj is not None:
                result = self.generate_dict_from_row(slc_obj)
            else:
                raise DatabaseException(self.OBJECT_NOT_FOUND.format("Slice", slc_id))
        except Exception as e:
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e
        return result

    def add_reservation(self, *, slc_guid: str, rsv_resid: str, rsv_category: int, rsv_state: int,
                        rsv_pending: int, rsv_joining: int, properties, lease_start: datetime = None,
                        lease_end: datetime = None, rsv_graph_node_id: str = None, oidc_claim_sub: str = None,
                        email: str = None, project_id: str = None, site: str = None, rsv_type: str = None,
                        components: List[Tuple[str, str, str]] = None, host: str = None, ip_subnet: str = None):
        """
        Add a reservation
        @param slc_guid slice guid
        @param rsv_resid reservation guid
        @param rsv_category category
        @param rsv_state state
        @param rsv_pending pending state
        @param rsv_joining join state
        @param properties pickled instance
        @param lease_start Lease start time
        @param lease_end Lease end time
        @param rsv_graph_node_id graph id
        @param oidc_claim_sub OIDC Sub claim
        @param email User email
        @param project_id Project id
        @param site site
        @param rsv_type reservation type
        @param components list of components
        @param host host
        @param ip_subnet ip_subnet
        """
        session = self.get_session()
        try:
            slc_id = self.get_slc_id_by_slice_id(slice_id=slc_guid)
            rsv_obj = Reservations(rsv_slc_id=slc_id, rsv_resid=rsv_resid, rsv_category=rsv_category,
                                   rsv_state=rsv_state, rsv_pending=rsv_pending, rsv_joining=rsv_joining,
                                   lease_start=lease_start, lease_end=lease_end,
                                   properties=properties, oidc_claim_sub=oidc_claim_sub, email=email,
                                   project_id=project_id, site=site, rsv_type=rsv_type, host=host, ip_subnet=ip_subnet)
            if rsv_graph_node_id is not None:
                rsv_obj.rsv_graph_node_id = rsv_graph_node_id

            # Create Component Mapping for the Network Service reservation
            if components:
                for node_id, cid, bdf in components:
                    mapping = Components(node_id=node_id, reservation=rsv_obj,
                                         component=cid, bdf=bdf)
                    session.add(mapping)

            session.add(rsv_obj)
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def update_reservation(self, *, slc_guid: str, rsv_resid: str, rsv_category: int, rsv_state: int,
                           rsv_pending: int, rsv_joining: int, properties, lease_start: datetime = None,
                           lease_end: datetime = None, rsv_graph_node_id: str = None, site: str = None,
                           rsv_type: str = None, components: List[Tuple[str, str, str]] = None,
                           host: str = None, ip_subnet: str = None):
        """
        Update a reservation
        @param slc_guid slice guid
        @param rsv_resid reservation guid
        @param rsv_category category
        @param rsv_state state
        @param rsv_pending pending state
        @param rsv_joining join state
        @param properties pickled instance
        @param lease_start Lease start time
        @param lease_end Lease end time
        @param rsv_graph_node_id graph id
        @param site site
        @param rsv_type reservation type
        @param components list of components
        @param ip_subnet ip subnet
        @param host host
        """
        session = self.get_session()
        try:
            rsv_obj = session.query(Reservations).filter_by(rsv_resid=rsv_resid).one()
            if rsv_obj is not None:
                rsv_obj.rsv_category = rsv_category
                rsv_obj.rsv_state = rsv_state
                rsv_obj.rsv_pending = rsv_pending
                rsv_obj.rsv_joining = rsv_joining
                rsv_obj.properties = properties
                rsv_obj.lease_end = lease_end
                rsv_obj.lease_start = lease_start
                if host:
                    rsv_obj.host = host
                if ip_subnet:
                    rsv_obj.ip_subnet = ip_subnet
                if site is not None:
                    rsv_obj.site = site
                if rsv_graph_node_id is not None:
                    rsv_obj.rsv_graph_node_id = rsv_graph_node_id
                if rsv_type is not None:
                    rsv_obj.rsv_type = rsv_type

                # Update components records for the reservation
                if components and len(components):
                    existing = session.query(Components).filter(Components.reservation_id == rsv_obj.rsv_id).all()
                    existing_components = {(c.node_id, c.component, c.bdf) for c in existing}

                    # Identify new string values
                    added_comps = set(components) - existing_components

                    # Identify outdated string values
                    removed_comps = existing_components - set(components)

                    # Remove outdated comps
                    for node_id, cid, bdf in removed_comps:
                        comp_to_remove = next(
                            (comp for comp in existing if comp.component == cid and
                             comp.node_id == node_id and comp.bdf == bdf),
                            None)
                        if comp_to_remove:
                            session.delete(comp_to_remove)

                    # Add new comps
                    for node_id, cid, bdf in added_comps:
                        new_mapping = Components(node_id=node_id, reservation=rsv_obj,
                                                 component=cid, bdf=bdf)
                        session.add(new_mapping)

            else:
                raise DatabaseException(self.OBJECT_NOT_FOUND.format("Reservation", rsv_resid))
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def remove_reservation(self, *, rsv_resid: str):
        """
        Remove a reservation
        @param rsv_resid reservation guid
        """
        session = self.get_session()
        try:
            #session.query(Reservations).filter_by(rsv_resid=rsv_resid).delete()
            reservation = session.query(Reservations).filter_by(rsv_resid=rsv_resid).one_or_none()

            if reservation:
                # Delete associated Components records
                mappings = session.query(Components).filter(Components.reservation_id == reservation.rsv_id).all()
                for mapping in mappings:
                    session.delete(mapping)

            session.delete(reservation)
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def create_reservation_filter(self, *, slice_id: str = None, graph_node_id: str = None, project_id: str = None,
                                  email: str = None, oidc_sub: str = None, rid: str = None, site: str = None,
                                  ip_subnet: str = None, host: str = None) -> dict:

        filter_dict = {}
        if slice_id is not None:
            slc_id = self.get_slc_id_by_slice_id(slice_id=slice_id)
            filter_dict['rsv_slc_id'] = slc_id
        if graph_node_id is not None:
            filter_dict['rsv_graph_node_id'] = graph_node_id
        if project_id is not None:
            filter_dict['project_id'] = project_id
        if email is not None:
            filter_dict['email'] = email
        if oidc_sub is not None:
            filter_dict['oidc_claim_sub'] = oidc_sub
        if rid is not None:
            filter_dict['rsv_resid'] = rid
        if site is not None:
            filter_dict['site'] = site
        if ip_subnet:
            filter_dict['ip_subnet'] = ip_subnet
        if host:
            filter_dict['host'] = host

        return filter_dict

    def get_reservations(self, *, slice_id: str = None, graph_node_id: str = None, project_id: str = None,
                         email: str = None, oidc_sub: str = None, rid: str = None, states: list[int] = None,
                         category: list[int] = None, site: str = None, rsv_type: list[str] = None,
                         start: datetime = None, end: datetime = None, ip_subnet: str = None,
                         host: str = None) -> List[dict]:
        """
        Get Reservations for an actor
        @param slice_id slice id
        @param graph_node_id graph node id
        @param project_id project id
        @param email email
        @param oidc_sub oidc sub
        @param rid reservation id
        @param states reservation state
        @param category reservation category
        @param site site name
        @param rsv_type rsv_type
        @param start search for slivers with lease_end_time after start
        @param end search for slivers with lease_end_time before end
        @param ip_subnet ip subnet
        @param host host

        @return list of reservations
        """
        result = []
        session = self.get_session()
        try:
            filter_dict = self.create_reservation_filter(slice_id=slice_id, graph_node_id=graph_node_id,
                                                         project_id=project_id, email=email, oidc_sub=oidc_sub,
                                                         rid=rid, site=site, ip_subnet=ip_subnet, host=host)
            rows = session.query(Reservations).filter_by(**filter_dict)

            if rsv_type is not None:
                rows = rows.filter(Reservations.rsv_type.in_(rsv_type))

            if states is not None:
                rows = rows.filter(Reservations.rsv_state.in_(states))

            if category is not None:
                rows = rows.filter(Reservations.rsv_category.in_(category))

            # Ensure start and end are datetime objects
            if start and isinstance(start, str):
                start = datetime.fromisoformat(start)
            if end and isinstance(end, str):
                end = datetime.fromisoformat(end)

            # Construct filter condition for lease_end within the given time range
            if start is not None or end is not None:
                lease_end_filter = True  # Initialize with True to avoid NoneType comparison
                if start is not None and end is not None:
                    lease_end_filter = or_(
                        and_(start <= Reservations.lease_end, Reservations.lease_end <= end),
                        and_(start <= Reservations.lease_start, Reservations.lease_start <= end),
                        and_(Reservations.lease_start <= start, Reservations.lease_end >= end)
                    )
                elif start is not None:
                    lease_end_filter = start <= Reservations.lease_end
                elif end is not None:
                    lease_end_filter = Reservations.lease_end <= end

                rows = rows.filter(lease_end_filter)

            for row in rows.all():
                result.append(self.generate_dict_from_row(row=row))
        except Exception as e:
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e
        return result

    def get_components(self, *, node_id: str, states: list[int], rsv_type: list[str], component: str = None,
                       bdf: str = None, start: datetime = None, end: datetime = None,
                       excludes: List[str] = None) -> Dict[str, List[str]]:
        """
        Returns components matching the search criteria
        @param node_id: Worker Node ID to which components belong
        @param states: list of states used to find reservations
        @param rsv_type: type of reservations
        @param component: component name
        @param bdf: Component's PCI address
        @param start: start time
        @param end: end time
        @param excludes: list of the reservations ids to exclude

        NOTE# For P4 switches; node_id=node+renc-p4-sw  component=ip+192.168.11.8 bdf=p1

        @return Dictionary with component name as the key and value as list of associated PCI addresses in use.
        """
        result = {}
        session = self.get_session()
        try:
            lease_end_filter = True  # Initialize with True to avoid NoneType comparison
            # Construct filter condition for lease_end within the given time range
            if start is not None or end is not None:
                if start is not None and end is not None:
                    lease_end_filter = or_(
                        and_(start <= Reservations.lease_end, Reservations.lease_end <= end),
                        and_(start <= Reservations.lease_start, Reservations.lease_start <= end),
                        and_(Reservations.lease_start <= start, Reservations.lease_end >= end)
                    )
                elif start is not None:
                    lease_end_filter = start <= Reservations.lease_end
                elif end is not None:
                    lease_end_filter = Reservations.lease_end <= end

            # Query to retrieve Components based on specific Reservation types and states
            rows = (
                session.query(Components)
                    .join(Reservations, Components.reservation_id == Reservations.rsv_id)
                    .filter(Reservations.rsv_type.in_(rsv_type))
                    .filter(Reservations.rsv_state.in_(states))
                    .filter(lease_end_filter)
                    .filter(Components.node_id == node_id)
                    .options(joinedload(Components.reservation))
            )

            # Add excludes filter if excludes list is not None and not empty
            if excludes:
                rows = rows.filter(Reservations.rsv_resid.notin_(excludes))

            # Query Component records for reservations in the specified state and owner with the target string
            if component is not None and bdf is not None:
                rows = rows.filter(Components.component == component, Components.bdf == bdf)
            elif component is not None:
                rows = rows.filter(Components.component == component)
            elif bdf is not None:
                rows = rows.filter(Components.bdf == bdf)

            for row in rows.all():
                if row.component not in result:
                    result[row.component] = []
                if row.bdf not in result[row.component]:
                    result[row.component].append(row.bdf)
        except Exception as e:
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e
        return result

    def get_reservations_by_rids(self, *, rsv_resid_list: list) -> list:
        """
        Get Reservations for an actor by reservation ids
        @param act_id actor id
        @param rsv_resid_list reservation guid list
        @return list of reservations
        """
        result = []
        session = self.get_session()
        try:
            for row in session.query(Reservations).filter(Reservations.rsv_resid.in_(rsv_resid_list)).all():
                result.append(self.generate_dict_from_row(row=row))
        except Exception as e:
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e
        return result

    def add_proxy(self, *, act_id: int, prx_name: str, properties):
        """
        Add a proxy
        @param act_id actor id
        @param prx_name proxy name
        @param properties pickled instance
        """
        session = self.get_session()
        try:
            prx_obj = Proxies(prx_act_id=act_id, prx_name=prx_name, properties=properties)
            session.add(prx_obj)
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def update_proxy(self, *, act_id: int, prx_name: str, properties):
        """
        Update a proxy
        @param act_id actor id
        @param prx_name proxy name
        @param properties pickled instance
        """
        session = self.get_session()
        try:
            prx_obj = session.query(Proxies).filter_by(prx_act_id=act_id).filter(
                Proxies.prx_name == prx_name).one_or_none()
            if prx_obj is not None:
                prx_obj.properties = properties
            else:
                raise DatabaseException(self.OBJECT_NOT_FOUND.format("Proxy", prx_name))
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def remove_proxy(self, *, act_id: int, prx_name: str):
        """
        Remove a proxy
        @param act_id actor id
        @param prx_name proxy name
        """
        session = self.get_session()
        try:
            session.query(Proxies).filter(Proxies.prx_act_id == act_id).filter(Proxies.prx_name == prx_name).delete()
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def get_proxies(self, *, act_id: int) -> list:
        """
        Get Proxies
        @param act_id actor id
        """
        result = []
        session = self.get_session()
        try:
            for row in session.query(Proxies).filter_by(prx_act_id=act_id).all():
                result.append(self.generate_dict_from_row(row=row))
        except Exception as e:
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e
        return result

    def add_config_mapping(self, *, act_id: int, cfgm_type: str, properties):
        """
        Add handlers mapping
        @param act_id actor id
        @param cfgm_type type
        @param properties properties
        """
        session = self.get_session()
        try:
            cfg_obj = ConfigMappings(cfgm_act_id=act_id, cfgm_type=cfgm_type, properties=properties)
            session.add(cfg_obj)
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def update_config_mapping(self, *, act_id: int, cfgm_type: str, properties):
        """
        Update handlers mapping
        @param act_id actor id
        @param cfgm_type type
        @param properties properties
        """
        session = self.get_session()
        try:
            cfg_obj = session.query(ConfigMappings).filter(ConfigMappings.cfgm_act_id == act_id).filter(
                ConfigMappings.cfgm_type == cfgm_type).one_or_none()
            if cfg_obj is not None:
                cfg_obj.properties = properties
            else:
                raise DatabaseException(self.OBJECT_NOT_FOUND.format("Config Mapping", cfgm_type))
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def remove_config_mapping(self, *, cfgm_type: str):
        """
        Remove handlers mapping
        @param cfgm_type handlers mapping type
        """
        session = self.get_session()
        try:
            session.query(ConfigMappings).filter_by(cfgm_type=cfgm_type).delete()
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def get_config_mappings(self, *, act_id: int) -> list:
        """
        Get Config Mappings
        @param act_id actor id
        @retur list of mappings
        """
        result = []
        session = self.get_session()
        try:
            for row in session.query(ConfigMappings).filter_by(cfgm_act_id=act_id).all():
                result.append(self.generate_dict_from_row(row=row))
        except Exception as e:
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e
        return result

    def add_client(self, *, act_id: int, clt_name: str, clt_guid: str, properties):
        """
        Add a client
        @param act_id actor id
        @param clt_name client name
        @param clt_guid client guid
        @param properties pickled instance
        """
        session = self.get_session()
        try:
            clt_obj = Clients(clt_act_id=act_id, clt_name=clt_name, clt_guid=clt_guid, properties=properties)
            session.add(clt_obj)
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def update_client(self, *, act_id: int, clt_name: str, properties):
        """
        Update a client
        @param act_id actor id
        @param clt_name client name
        @param properties pickled instance
        """
        session = self.get_session()
        try:
            clt_obj = session.query(Clients).filter(Clients.clt_act_id == act_id).filter(
                Clients.clt_name == clt_name).one_or_none()
            if clt_obj is not None:
                clt_obj.properties = properties
            else:
                raise DatabaseException(self.OBJECT_NOT_FOUND.format("Client", clt_name))
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def remove_client_by_name(self, *, act_id: int, clt_name: str):
        """
        Remove a client by name
        @param act_id actor id
        @param clt_name client name
        """
        session = self.get_session()
        try:
            session.query(Clients).filter(Clients.clt_act_id == act_id).filter(Clients.clt_name == clt_name).delete()
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def remove_client_by_guid(self, *, act_id: int, clt_guid: str):
        """
        Remove client by guid
        @param act_id actor id
        @param clt_guid client guid
        """
        session = self.get_session()
        try:
            session.query(Clients).filter(Clients.clt_act_id == act_id).filter(Clients.clt_guid == clt_guid).delete()
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def get_client_by_guid(self, *, clt_guid: str) -> dict:
        """
        Get Client by name
        @param act_id actor id
        @param clt_guid client guid
        @return client
        """
        result = None
        session = self.get_session()
        try:
            clt_obj = session.query(Clients).filter_by(clt_guid=clt_guid).one_or_none()
            if clt_obj is None:
                raise DatabaseException(self.OBJECT_NOT_FOUND.format("Client", clt_guid))
            result = self.generate_dict_from_row(clt_obj)
        except Exception as e:
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e
        return result

    def get_clients(self) -> list:
        """
        Get Clients
        @param act_id actor id
        @return client list
        """
        result = []
        session = self.get_session()
        try:
            for row in session.query(Clients).all():
                result.append(self.generate_dict_from_row(row=row))
        except Exception as e:
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e
        return result

    def add_unit(self, *, slc_guid: str, rsv_resid: str, unt_uid: str, unt_unt_id: int,
                 unt_state: int, properties):
        """
        Add Unit
        @param act_id actor id
        @param slc_guid slice guid
        @param rsv_resid reservation id
        @param unt_uid unit id
        @param unt_unt_id parent unit id
        @param unt_state unit state
        @param properties properties
        """
        session = self.get_session()
        try:
            slc_id = self.get_slc_id_by_slice_id(slice_id=slc_guid)
            rsv_id = self.get_rsv_id_by_reservation_id(reservation_id=rsv_resid)

            unt_obj = Units(unt_uid=unt_uid,
                            unt_slc_id=slc_id,
                            unt_rsv_id=rsv_id,
                            unt_state=unt_state, properties=properties)
            if unt_unt_id is not None:
                unt_obj.unt_unt_id = unt_unt_id
            session.add(unt_obj)
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def get_unit(self, *, unt_uid: str) -> dict or None:
        """
        Get Unit
        @param act_id actor id
        @param unt_uid unit guid
        @return Unit dict
        """
        result = None
        session = self.get_session()
        try:
            unt_obj = session.query(Units).filter_by(unt_uid=unt_uid).one_or_none()
            if unt_obj is None:
                self.logger.debug("Unit with guid {} not found".format(unt_uid))
                return result
            result = self.generate_dict_from_row(unt_obj)
        except Exception as e:
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e
        return result

    def get_units(self, *, rsv_resid: str):
        """
        Get Units
        @param act_id actor id
        @param rsv_resid reservation guid
        @return Unit list
        """
        result = []
        session = self.get_session()
        try:
            rsv_obj = self.get_reservations(rid=rsv_resid)[0]
            if rsv_obj is None:
                raise DatabaseException(self.OBJECT_NOT_FOUND.format("Reservation", rsv_resid))

            for row in session.query(Units).filter_by(unt_rsv_id=rsv_obj['rsv_id']).all():
                result.append(self.generate_dict_from_row(row=row))
        except Exception as e:
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e
        return result

    def remove_unit(self, *, unt_uid: str):
        """
        Remove a unit
        @param unt_uid unit id
        """
        session = self.get_session()
        try:
            session.query(Units).filter_by(unt_uid=unt_uid).delete()
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def update_unit(self, *, unt_uid: str, properties):
        """
        Add Unit
        @param unt_uid unit id
        @param properties properties
        """
        result = None
        session = self.get_session()
        try:
            unt_obj = session.query(Units).filter_by(unt_uid=unt_uid).one_or_none()
            if unt_obj is None:
                raise DatabaseException(self.OBJECT_NOT_FOUND.format("Unit", unt_uid))
            unt_obj.properties = properties
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e
        return result

    def add_delegation(self, *, slice_id: str, dlg_graph_id: str, dlg_state: int, properties, site: str = None):
        """
        Add delegation
        @param slice_id slice id
        @param dlg_graph_id graph id
        @param dlg_state state
        @param properties properties
        @param site site
        """
        session = self.get_session()
        try:
            slc_id = self.get_slc_id_by_slice_id(slice_id=slice_id)
            dlg_obj = Delegations(dlg_slc_id=slc_id, dlg_graph_id=dlg_graph_id,
                                  dlg_state=dlg_state, properties=properties, site=site)
            session.add(dlg_obj)
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def update_delegation(self, *, dlg_graph_id: str, dlg_state: int, properties, site: str = None):
        """
        Update delegation
        @param dlg_graph_id graph id
        @param dlg_state state
        @param properties properties
        @param site site
        """
        session = self.get_session()
        try:
            dlg_obj = session.query(Delegations).filter_by(dlg_graph_id=dlg_graph_id).one_or_none()
            if dlg_obj is not None:
                dlg_obj.dlg_state = dlg_state
                dlg_obj.properties = properties
                if site is not None:
                    dlg_obj.site = site
            else:
                raise DatabaseException(self.OBJECT_NOT_FOUND.format("Delegation", dlg_graph_id))
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def remove_delegation(self, *, dlg_graph_id: str):
        """
        Remove delegation
        @param dlg_graph_id graph id
        """
        session = self.get_session()
        try:
            session.query(Delegations).filter_by(dlg_graph_id=dlg_graph_id).delete()
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def get_slc_id_by_slice_id(self, *, slice_id: str) -> int:
        slices = self.get_slices(slice_id=slice_id)
        if slices is None or len(slices) == 0:
            raise DatabaseException(self.OBJECT_NOT_FOUND.format("Slice", slice_id))
        return slices[0]['slc_id']

    def get_rsv_id_by_reservation_id(self, *, reservation_id: str) -> int:
        reservations = self.get_reservations(rid=reservation_id)
        if reservations is None or len(reservations) == 0:
            raise DatabaseException(self.OBJECT_NOT_FOUND.format("Reservation", reservation_id))
        return reservations[0]['rsv_id']

    def get_delegations(self, *, slc_guid: str = None, states: List[int] = None) -> List[dict]:
        """
        Get delegations
        @param slc_guid slice guid
        @param states delegation state
        @param list of delegations
        """
        result = []
        session = self.get_session()
        try:
            slc_id = self.get_slc_id_by_slice_id(slice_id=slc_guid)
            rows = session.query(Delegations)
            if slc_guid is not None:
                rows = rows.filter(Delegations.dlg_slc_id == slc_id)
            if states is not None:
                rows = rows.filter(Delegations.dlg_state.in_(states))
            for row in rows.all():
                result.append(self.generate_dict_from_row(row=row))
        except Exception as e:
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e
        return result

    def get_delegation(self, *, dlg_graph_id: str) -> dict:
        """
        Get delegation
        @param dlg_graph_id delegation graph id
        @return delegation
        """
        result = {}
        session = self.get_session()
        try:
            dlg_obj = session.query(Delegations).filter_by(dlg_graph_id=dlg_graph_id).one_or_none()
            if dlg_obj is not None:
                result = self.generate_dict_from_row(dlg_obj)
            else:
                raise DatabaseException(self.OBJECT_NOT_FOUND.format("Delegation", dlg_graph_id))
        except Exception as e:
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e
        return result

    def add_site(self, *, site_name: str, state: int, properties):
        """
        Add a site
        @param site_name Site Name
        @param state state
        @param properties pickled instance
        """
        session = self.get_session()
        try:
            site_obj = Sites(name=site_name, state=state, properties=properties)
            session.add(site_obj)
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def update_site(self, *, site_name: str, state: int, properties):
        """
        Update a site
        @param site_name Site Name
        @param state state
        @param properties pickled instance
        """
        session = self.get_session()
        try:
            site_obj = session.query(Sites).filter_by(name=site_name).one_or_none()
            if site_obj is not None:
                site_obj.properties = properties
                site_obj.state = state
            else:
                raise DatabaseException(self.OBJECT_NOT_FOUND.format("Sites", site_name))
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def remove_site(self, *, site_name: str):
        """
        Remove a proxy
        @param site_name site_name
        """
        session = self.get_session()
        try:
            session.query(Sites).filter_by(name=site_name).delete()
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def get_sites(self) -> list:
        """
        Get Sites
=        """
        result = []
        session = self.get_session()
        try:
            for row in session.query(Sites).all():
                result.append(self.generate_dict_from_row(row=row))
        except Exception as e:
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e
        return result

    def get_site(self, *, site_name: str) -> list:
        """
        Get Sites
=        """
        result = []
        session = self.get_session()
        try:
            for row in session.query(Sites).filter_by(name=site_name).all():
                result.append(self.generate_dict_from_row(row=row))
        except Exception as e:
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e
        return result

    @staticmethod
    def generate_dict_from_row(row):
        d = row.__dict__.copy()
        for k in row.__dict__:
            if d[k] is None:
                d.pop(k)
        return d

    def add_poa(self, *, poa_guid: str, properties, email: str, project_id: str, sliver_id: str, slice_id: str,
                state: int):
        """
        Add a POA
        @param poa_guid POA id
        @param properties pickled instance
        @param project_id User OIDC Sub
        @param email User Email
        @param sliver_id Sliver Id
        @param slice_id slice id
        @param state state
        """
        session = self.get_session()
        try:
            poa_obj = Poas(poa_guid=poa_guid, project_id=project_id, email=email, sliver_id=sliver_id,
                           slice_id=slice_id, state=state,last_update_time=datetime.now(timezone.utc),
                           properties=properties)
            session.add(poa_obj)
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def update_poa(self, *, poa_guid: str, properties, state: int, email: str = None, project_id: str = None,
                   sliver_id: str = None):
        """
        Update a POA
        @param poa_guid POA id
        @param properties pickled instance
        @param state state
        @param project_id User OIDC Sub
        @param email User Email
        @param sliver_id Sliver Id
        """
        session = self.get_session()
        try:
            poa_obj = session.query(Poas).filter_by(poa_guid=poa_guid).one_or_none()
            if poa_obj is not None:
                poa_obj.properties = properties
                if email is not None:
                    poa_obj.email = email
                if project_id is not None:
                    poa_obj.project_id = project_id
                if sliver_id is not None:
                    poa_obj.sliver_id = sliver_id
                poa_obj.last_update_time = datetime.now(timezone.utc)
                poa_obj.state = state
            else:
                raise DatabaseException(self.OBJECT_NOT_FOUND.format("Poa", poa_guid))
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def remove_poa(self, *, poa_guid: str):
        """
        Remove POA
        @param poa_guid POA Guid
        """
        session = self.get_session()
        try:
            session.query(Poas).filter_by(poa_guid=poa_guid).delete()
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    @staticmethod
    def create_poa_filter(*, poa_guid: str = None, sliver_id: str = None, project_id: str = None,
                          email: str = None, slice_id: str = None) -> dict:

        filter_dict = {}
        if poa_guid is not None:
            filter_dict['poa_guid'] = poa_guid
        if sliver_id is not None:
            filter_dict['sliver_id'] = str(sliver_id)
        if slice_id is not None:
            filter_dict['slice_id'] = str(slice_id)
        if project_id is not None:
            filter_dict['project_id'] = project_id
        if email is not None:
            filter_dict['email'] = email
        return filter_dict

    def get_poas(self, *, poa_guid: str = None, project_id: str = None, email: str = None, sliver_id: str = None,
                 slice_id: str, limit: int = None, offset: int = None, last_update_time: datetime = None,
                 states: list[int] = None) -> List[dict]:
        """
        Get slices for an actor
        @param poa_guid POA Guid
        @param project_id project id
        @param email email
        @param limit limit
        @param offset offset
        @param sliver_id Sliver Id
        @param slice_id Slice Id
        @param last_update_time Last Update Time
        @param states
        @return list of POAs
        """
        result = []
        session = self.get_session()
        try:
            filter_dict = self.create_poa_filter(poa_guid=poa_guid, project_id=project_id, email=email,
                                                 sliver_id=sliver_id, slice_id=slice_id)

            rows = session.query(Poas).filter_by(**filter_dict)

            if last_update_time is not None:
                rows = rows.filter(Poas.last_update_time < last_update_time)

            if states is not None:
                rows = rows.filter(Poas.state.in_(states))

            rows = rows.order_by(desc(Poas.last_update_time))

            if offset is not None and limit is not None:
                rows = rows.offset(offset).limit(limit)

            for row in rows.all():
                result.append(self.generate_dict_from_row(row=row))
        except Exception as e:
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e
        return result

    def increment_metrics(self, *, project_id: str, user_id: str, slice_count: int = 1):
        """
        Add or Update Metrics
        @param project_id: project_id
        @param user_id: user_id
        @param slice_count: slice_count
        """
        session = self.get_session()
        try:
            metric_obj = session.query(Metrics).filter_by(project_id=project_id, user_id=user_id).one_or_none()
            if not metric_obj:
                metric_obj = Metrics(project_id=project_id, user_id=user_id, slice_count=slice_count)
                session.add(metric_obj)
            else:
                metric_obj.slice_count += slice_count
            session.commit()
        except Exception as e:
            session.rollback()
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e

    def get_metrics(self, *, project_id: str = None, user_id: str = None, excluded_projects: List[str] = None) -> list:
        """
        Get Metric count
        @param project_id: project_id
        @param user_id: user_id
        @param excluded_projects: excluded_projects
        @return list of metrics
        """
        result = []
        session = self.get_session()
        try:
            filter_criteria = True
            # Construct filter condition
            if project_id and user_id:
                filter_criteria = and_(Metrics.project_id == project_id, Metrics.user_id == user_id)
            elif project_id is not None:
                filter_criteria = and_(Metrics.project_id == project_id)
            elif user_id is not None:
                filter_criteria = and_(Metrics.user_id == user_id)

            if excluded_projects:
                filter_criteria = and_(Metrics.project_id.notin_(excluded_projects))

            rows = session.query(Metrics).filter(filter_criteria).all()

            for r in rows:
                result.append(self.generate_dict_from_row(row=r))
            return result
        except Exception as e:
            self.logger.error(Constants.EXCEPTION_OCCURRED.format(e))
            raise e


def test():
    logger = logging.getLogger('PsqlDatabase')
    db = PsqlDatabase(user='fabric', password='fabric', database='am', db_host='127.0.0.1:5432', logger=logger)
    db.create_db()
    db.reset_db()

    # Actor Operations
    prop = {'abc':'def'}

    db.add_actor(name="test-actor1", guid="12345", act_type=1, properties=pickle.dumps(prop))
    print("All actors: {}".format(db.get_actors()))
    print("one actor {}".format(db.get_actor(name="test-actor1")))
    prop['pqr'] = 'stu'
    db.update_actor(name="test-actor1", properties=pickle.dumps(prop))
    print("actor after update {}".format(db.get_actor(name="test-actor1")))
    db.remove_actor(name="test-actor1")
    print("actors after delete: {}".format(db.get_actors()))

    # Miscellaneous operations
    db.add_miscellaneous(name='time', properties=prop)
    print(db.get_miscellaneous(name='time'))

    # Manager Object operations
    db.add_actor(name="test-actor1", guid="12345", act_type=1, properties=pickle.dumps(prop))
    print("All actors: {}".format(db.get_actors()))
    actor = db.get_actor(name="test-actor1")
    actor_id = actor['act_id']
    db.add_manager_object(manager_key="mgr-1234", properties=prop, act_id=actor_id)
    db.add_manager_object(manager_key="mgr-1235", properties=prop)
    print(db.get_manager_objects())
    print("Get specific MO {}".format(db.get_manager_object(mo_key="mgr-1234")))
    print("Get MO by actor {}".format(db.get_manager_objects(act_id=actor_id)))
    print("Get MO with no actor {}".format(db.get_manager_containers()))
    db.remove_manager_object(manager_key="mgr-1235")
    print("After removing MO by key {}".format(db.get_manager_objects()))
    db.remove_manager_object_by_actor(act_id=actor_id)
    print("After removing MO by actor {}".format(db.get_manager_objects()))

    # Slice operations
    db.add_slice(slc_guid="1234", slc_name="test-slice", slc_type=1, slc_resource_type="def",
                 properties=pickle.dumps(prop), lease_start=datetime.now(timezone.utc),
                 lease_end=datetime.now(timezone.utc), project_id="12345", slc_state=1)
    print("Slices after add slice {}".format(db.get_slices()))
    prop['fabric'] = 'testbed'
    db.update_slice(slc_guid="1234", slc_name="test-slice", slc_type=1, slc_resource_type="def",
                    properties=pickle.dumps(prop), lease_start=datetime.now(timezone.utc),
                    lease_end=datetime.now(timezone.utc), slc_state=1)
    print("Get slice after update {}".format(db.get_slices(slice_id="1234")))

    print("Get slice by type after update {}".format(db.get_slices(slc_type=[1])))

    db.remove_slice(slc_guid="1234")
    print("Get all slices {} after delete".format(db.get_slices()))

    # Reservation operations
    db.add_slice(slc_guid="1234", slc_name="test-slice", slc_type=1, slc_resource_type="def",
                 properties=pickle.dumps(prop), lease_start=datetime.now(timezone.utc),
                 lease_end=datetime.now(timezone.utc), slc_state=1)
    print("Get all slices {} after add".format(db.get_slices()))
    prop['rsv'] = 'test'
    db.add_reservation(slc_guid="1234", rsv_resid="rsv_567", rsv_category=1, rsv_state=2,
                       rsv_pending=3, rsv_joining=4, properties=pickle.dumps(prop))
    print("Reservations after add {}".format(db.get_reservations()))
    prop['update'] = 'test'
    db.update_reservation(slc_guid="1234", rsv_resid="rsv_567", rsv_category=1, rsv_state=2,
                          rsv_pending=3, rsv_joining=5, properties=pickle.dumps(prop))
    print("Reservations after update {}".format(db.get_reservations()))
    print("Reservations after by state {}".format(db.get_reservations(slice_id="1234", states=[2])))

    print("Reservations after by rid {}".format(db.get_reservations(rid="rsv_567")))
    print("Reservations after by rid list {}".format(db.get_reservations_by_rids(rsv_resid_list=["rsv_567"])))

    db.remove_reservation(rsv_resid="rsv_567")
    print("Reservations after delete {}".format(db.get_reservations()))

    # Proxy operations
    db.add_proxy(act_id=actor_id, prx_name="prx-123", properties=pickle.dumps(prop))
    print("Proxies after add {}".format(db.get_proxies(act_id=actor_id)))
    prop['prx-update'] = 'fly'
    db.update_proxy(act_id=actor_id, prx_name="prx-123", properties=pickle.dumps(prop))
    print("Proxies after update {}".format(db.get_proxies(act_id=actor_id)))
    db.remove_proxy(act_id=actor_id, prx_name="prx-123")
    print("Proxies after remove {}".format(db.get_proxies(act_id=actor_id)))

    # Config Mapping operations
    db.add_config_mapping(act_id=actor_id, cfgm_type="cfg-1", properties=pickle.dumps(prop))
    print("Config Mappings after add {}".format(db.get_config_mappings(act_id=actor_id)))
    prop['cfg-update'] = 'done'
    db.update_config_mapping(act_id=actor_id, cfgm_type="cfg-1", properties=pickle.dumps(prop))
    db.remove_config_mapping(cfgm_type="cfg-1")
    print("Config Mappings after remove {}".format(db.get_config_mappings(act_id=actor_id)))

    # Client operations
    db.add_client(act_id=actor_id, clt_name="clt-test", clt_guid="clt-1234", properties=pickle.dumps(prop))
    print("Clients after add {}".format(db.get_clients()))
    print("Clients after by guid {}".format(db.get_client_by_guid(clt_guid="clt-1234")))
    prop['clt-up'] = 'done'
    db.update_client(act_id=actor_id, clt_name="clt-test", properties=pickle.dumps(prop))
    db.remove_client_by_guid(act_id=actor_id, clt_guid="clt-1234")
    print("Clients after remove {}".format(db.get_clients()))

    # Unit operations
    db.add_reservation(slc_guid="1234", rsv_resid="rsv_567", rsv_category=1, rsv_state=2,
                       rsv_pending=3, rsv_joining=4, properties=pickle.dumps(prop))

    db.add_unit(slc_guid="1234", rsv_resid="rsv_567", unt_uid="unt_123", unt_unt_id=None,
                unt_state=2, properties=pickle.dumps(prop))
    print("Unit after add {}".format(db.get_units(rsv_resid="rsv_567")))
    prop['update_unit'] = 'done'
    db.update_unit(unt_uid="unt_123", properties=pickle.dumps(prop))
    print("Unit after update {}".format(db.get_unit(unt_uid="unt_123")))
    db.remove_unit(unt_uid="unt_123")
    print("Unit after remove {}".format(db.get_units(rsv_resid="rsv_567")))

    db.remove_reservation(rsv_resid="rsv_567")
    db.reset_db()
    print("Get all actors after reset {}".format(db.get_actors()))


def test2():
    logger = logging.getLogger('PsqlDatabase')
    db = PsqlDatabase(user='fabric', password='fabric', database='broker', db_host='127.0.0.1:5432', logger=logger)
    from fabric_cf.actor.core.kernel.reservation_states import ReservationStates
    states = [ReservationStates.Active.value,
              ReservationStates.ActiveTicketed.value,
              ReservationStates.Ticketed.value,
              ReservationStates.Nascent.value]
    res = db.get_reservations(graph_node_id='node+renc-data-sw:ip+192.168.11.3-ns', states=states)
    print(f"All {len(res)}")
    for r in res:
        print(r['rsv_state'])
    print("Get all actors after reset {}".format(db.get_actors()))

def test3():
    logger = logging.getLogger('PsqlDatabase')
    db = PsqlDatabase(user='fabric', password='fabric', database='am', db_host='127.0.0.1:5432', logger=logger)
    db.create_db()
    db.reset_db()

    # Actor Operations
    prop = {'abc': 'def'}

    # Slice operations
    from fabric_cf.actor.core.kernel.slice_state_machine import SliceState
    db.add_slice(slc_guid="1234", slc_name="test-slice", slc_type=1, slc_resource_type="def",
                 properties=pickle.dumps(prop), lease_start=datetime.now(timezone.utc), lease_end=datetime.now(timezone.utc),
                 slc_state=SliceState.Closing.value, email="kthare10@email.unc.edu")

    db.add_slice(slc_guid="1234", slc_name="test-slice2", slc_type=1, slc_resource_type="def",
                 properties=pickle.dumps(prop), lease_start=datetime.now(timezone.utc), lease_end=datetime.now(timezone.utc),
                 slc_state=SliceState.Dead.value, email="kthare10@email.unc.edu")

    db.add_slice(slc_guid="1234", slc_name="test-slice3", slc_type=1, slc_resource_type="def",
                 properties=pickle.dumps(prop), lease_start=datetime.now(timezone.utc), lease_end=datetime.now(timezone.utc),
                 slc_state=SliceState.StableOK.value, email="kthare10@email.unc.edu")

    db.add_slice(slc_guid="1234", slc_name="test-slice4", slc_type=1, slc_resource_type="def",
                 properties=pickle.dumps(prop), lease_start=datetime.now(timezone.utc), lease_end=datetime.now(timezone.utc),
                 slc_state=SliceState.StableError.value, email="kthare10@email.unc.edu")

    db.add_slice(slc_guid="1234", slc_name="test-slice3", slc_type=1, slc_resource_type="def",
                 properties=pickle.dumps(prop), lease_start=datetime.now(timezone.utc), lease_end=datetime.now(timezone.utc),
                 slc_state=SliceState.Configuring.value, email="kthare10@email.unc.edu")

    ss = db.get_slices(states=[SliceState.Dead.value, SliceState.Closing.value],
                       email="kthare10@email.unc.edu")

    assert len(ss) == 2

    ss = db.get_slices(states=[SliceState.StableOK.value, SliceState.StableError.value],
                       email="kthare10@email.unc.edu")

    assert len(ss) == 2

    ss = db.get_slices(states=[SliceState.Configuring.value],
                       email="kthare10@email.unc.edu")

    assert len(ss) == 1


if __name__ == '__main__':
    test2()

    logger = logging.getLogger('PsqlDatabase')
    db = PsqlDatabase(user='fabric', password='fabric', database='orchestrator', db_host='127.0.0.1:5432',
                      logger=logger)
    #comps = db.get_components(node_id="HX7LQ53")
    ss = db.get_slices(updated_after=datetime(1970, 1, 1, tzinfo=timezone.utc))
    print(len(ss))



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)

from sqlalchemy import JSON, ForeignKey, LargeBinary, Index, TIMESTAMP, func
from sqlalchemy.orm import declarative_base
from sqlalchemy import Column, String, Integer, Sequence
from sqlalchemy.orm import relationship


Base = declarative_base()

FOREIGN_KEY_ACTOR_ID = 'Actors.act_id'
FOREIGN_KEY_SLICE_ID = 'Slices.slc_id'
FOREIGN_KEY_RESERVATION_ID = 'Reservations.rsv_id'


class Actors(Base):
    """
    Represents Actors Database Table
    """
    __tablename__ = 'Actors'
    act_id = Column(Integer, Sequence('act_id', start=1, increment=1), autoincrement=True, primary_key=True)
    act_name = Column(String, unique=True, nullable=False)
    act_guid = Column(String, nullable=False)
    act_type = Column(Integer, nullable=False)
    properties = Column(LargeBinary)


class Clients(Base):
    """
    Represents Clients Database Table
    """
    __tablename__ = 'Clients'
    clt_id = Column(Integer, Sequence('clt_id', start=1, increment=1), autoincrement=True, primary_key=True)
    clt_act_id = Column(Integer, ForeignKey(FOREIGN_KEY_ACTOR_ID))
    clt_name = Column(String, nullable=False)
    clt_guid = Column(String, nullable=False)
    properties = Column(LargeBinary)


class ConfigMappings(Base):
    """
    Represents ConfigMappings Database Table
    """
    __tablename__ = 'ConfigMappings'
    cfgm_id = Column(Integer, Sequence('cfgm_id', start=1, increment=1), autoincrement=True, primary_key=True)
    cfgm_act_id = Column(Integer, ForeignKey(FOREIGN_KEY_ACTOR_ID))
    cfgm_type = Column(String, nullable=False)
    properties = Column(LargeBinary)


class ManagerObjects(Base):
    """
    Represents ManagerObjects Database Table
    """
    __tablename__ = 'ManagerObjects'
    mo_id = Column(Integer, Sequence('mo_id', start=1, increment=1), autoincrement=True, primary_key=True)
    mo_key = Column(String, nullable=False, unique=True)
    mo_act_id = Column(Integer, ForeignKey(FOREIGN_KEY_ACTOR_ID))
    properties = Column(JSON)


class Miscellaneous(Base):
    """
    Represents Miscellaneous Database Table
    """
    __tablename__ = 'Miscellaneous'
    msc_id = Column(Integer, Sequence('msc_id', start=1, increment=1), autoincrement=True, primary_key=True)
    msc_path = Column(String, nullable=False, unique=True)
    properties = Column(JSON)


class Metrics(Base):
    """
    Represents Metrics Database Table
    """
    __tablename__ = 'Metrics'
    m_id = Column(Integer, Sequence('m_id', start=1, increment=1), autoincrement=True, primary_key=True)
    user_id = Column(String, nullable=False, index=True)
    project_id = Column(String, nullable=False, index=True)
    slice_count = Column(Integer, nullable=False)


class Proxies(Base):
    """
    Represents Proxies Database Table
    """
    __tablename__ = 'Proxies'
    prx_id = Column(Integer, Sequence('prx_id', start=1, increment=1), autoincrement=True, primary_key=True)
    prx_act_id = Column(Integer, ForeignKey(FOREIGN_KEY_ACTOR_ID))
    prx_name = Column(String)
    properties = Column(LargeBinary)


class Reservations(Base):
    """
    Represents Reservations Database Table
    """
    __tablename__ = 'Reservations'
    rsv_id = Column(Integer, Sequence('rsv_id', start=1, increment=1), autoincrement=True, primary_key=True)
    rsv_graph_node_id = Column(String, nullable=True, index=True)
    rsv_slc_id = Column(Integer, ForeignKey(FOREIGN_KEY_SLICE_ID), index=True)
    rsv_resid = Column(String, nullable=False, index=True)
    oidc_claim_sub = Column(String, nullable=True, index=True)
    host = Column(String, nullable=True, index=True)
    ip_subnet = Column(String, nullable=True, index=True)
    email = Column(String, nullable=True, index=True)
    project_id = Column(String, nullable=True, index=True)
    site = Column(String, nullable=True, index=True)
    rsv_type = Column(String, nullable=True, index=True)
    rsv_state = Column(Integer, nullable=False, index=True)
    rsv_category = Column(Integer, nullable=False)
    rsv_pending = Column(Integer, nullable=False)
    rsv_joining = Column(Integer, nullable=False)
    lease_start = Column(TIMESTAMP(timezone=True), nullable=True)
    lease_end = Column(TIMESTAMP(timezone=True), nullable=True)
    properties = Column(LargeBinary)
    components = relationship('Components', back_populates='reservation')

    Index('idx_slc_guid_resid', rsv_slc_id, rsv_resid)
    Index('idx_slc_guid_resid_email', rsv_slc_id, rsv_resid, email)
    Index('idx_resid_state', rsv_resid, rsv_state)
    Index('idx_slcid_state', rsv_slc_id, rsv_state)
    Index('idx_graph_id_res_id', rsv_graph_node_id, rsv_resid)
    Index('idx_host', host)
    Index('idx_ip_subnet', ip_subnet)


class Slices(Base):
    """
    Represents Slices Database Table
    """
    __tablename__ = 'Slices'
    slc_id = Column(Integer, Sequence('slc_id', start=1, increment=1), autoincrement=True, primary_key=True)
    slc_graph_id = Column(String, index=True)
    oidc_claim_sub = Column(String, index=True)
    email = Column(String, nullable=True, index=True)
    project_id = Column(String, nullable=True, index=True)
    slc_guid = Column(String, nullable=False, index=True)
    slc_name = Column(String, nullable=False, index=True)
    slc_state = Column(Integer, nullable=False, index=True)
    slc_type = Column(Integer, nullable=False, index=True)
    slc_resource_type = Column(String)
    lease_start = Column(TIMESTAMP(timezone=True), nullable=True)
    lease_end = Column(TIMESTAMP(timezone=True), nullable=True)
    properties = Column(LargeBinary)

    # New column for tracking last update time
    last_update_time = Column(TIMESTAMP(timezone=True), nullable=False, server_default=func.now(), onupdate=func.now())

    Index('idx_slc_guid_name', slc_guid, slc_name)
    Index('idx_slc_guid_name_email', slc_guid, slc_name, email)
    Index('idx_slc_guid_state', slc_guid, slc_state)


class Units(Base):
    """
    Represents Units Database Table
    """
    __tablename__ = 'Units'
    unt_id = Column(Integer, Sequence('unt_id', start=1, increment=1), autoincrement=True, primary_key=True)
    unt_uid = Column(String)
    unt_unt_id = Column(Integer, nullable=True)
    unt_slc_id = Column(Integer, ForeignKey(FOREIGN_KEY_SLICE_ID))
    unt_rsv_id = Column(Integer, ForeignKey(FOREIGN_KEY_RESERVATION_ID))
    unt_state = Column(Integer, nullable=False)
    properties = Column(LargeBinary)


class Delegations(Base):
    """
    Represents Delegations Database Table
    """
    __tablename__ = 'Delegations'
    dlg_id = Column(Integer, Sequence('dlg_id', start=1, increment=1), autoincrement=True, primary_key=True)
    dlg_slc_id = Column(Integer, ForeignKey(FOREIGN_KEY_SLICE_ID))
    site = Column(String, nullable=True, index=True)
    dlg_graph_id = Column(String, nullable=False)
    dlg_state = Column(Integer, nullable=False)
    properties = Column(LargeBinary)


class Sites(Base):
    """
    Represents Sites Database Table
    """
    __tablename__ = 'Sites'
    site_id = Column(Integer, Sequence('site_id', start=1, increment=1), autoincrement=True, primary_key=True)
    name = Column(String, index=True)
    state = Column(Integer, nullable=False)
    properties = Column(LargeBinary)


class Poas(Base):
    """
    Represents Poas Database Table
    """
    __tablename__ = 'Poas'
    poa_id = Column(Integer, Sequence('poa_id', start=1, increment=1), autoincrement=True, primary_key=True)
    poa_guid = Column(String, nullable=False, index=True)
    email = Column(String, nullable=True, index=True)
    project_id = Column(String, nullable=True, index=True)
    sliver_id = Column(String, nullable=True, index=True)
    state = Column(Integer, nullable=False, index=True)
    slice_id = Column(String, nullable=True, index=True)
    last_update_time = Column(TIMESTAMP(timezone=True), nullable=True)
    properties = Column(LargeBinary)

    Index('idx_poa_guid_email', poa_guid, email)
    Index('idx_poa_guid_project_id', poa_guid, project_id)
    Index('idx_poa_guid_sliver_id', poa_guid, sliver_id)
    Index('idx_poa_guid_email_sliver_id', poa_guid, email, sliver_id)
    Index('idx_poa_guid_email_project_id', poa_guid, email, project_id)


class Components(Base):
    __tablename__ = 'Components'
    reservation_id = Column(Integer, ForeignKey('Reservations.rsv_id'), primary_key=True)
    node_id = Column(String, primary_key=True, index=True)
    component = Column(String, primary_key=True, index=True)
    bdf = Column(String, primary_key=True, index=True)
    reservation = relationship('Reservations', back_populates='components')



#!/usr/bin/env python3
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import argparse
import logging
import traceback
import os
from datetime import datetime, timezone
from logging.handlers import RotatingFileHandler

from fim.slivers.interface_info import InterfaceSliver
from fim.slivers.network_node import NodeSliver
from fim.slivers.network_service import NetworkServiceSliver

from fabric_cf.actor.core.kernel.slice import SliceTypes, Slice
from fabric_cf.actor.core.plugins.db.actor_database import ActorDatabase
from fabric_cf.actor.core.container.globals import Globals, GlobalsSingleton
from fabric_cf.actor.core.policy.inventory_for_type import InventoryForType
from export.db_manager import DatabaseManager


LAST_EXPORT_FILE = "./last_export_time.txt"


class ExportScript:
    """
    CLI interface to fetch data from Postgres and insert it into the new SQLAlchemy database via DatabaseManager.
    """

    def __init__(self, src_user, src_password, src_db, src_host,
                 dest_user, dest_password, dest_db, dest_host, batch_size=1000):
        """
        Initializes connections to both source (Postgres) and destination (DatabaseManager).
        """
        self.logger = logging.getLogger("export")
        file_handler = RotatingFileHandler('/var/log/actor/export.log', backupCount=5, maxBytes=50000)
        logging.basicConfig(level=logging.INFO,
                            format="%(asctime)s [%(filename)s:%(lineno)d] [%(levelname)s] %(message)s",
                            handlers=[logging.StreamHandler(), file_handler])

        # Connect to the source Postgres database
        Globals.config_file = '/etc/fabric/actor/config/config.yaml'
        GlobalsSingleton.get().load_config()
        GlobalsSingleton.get().initialized = True

        self.src_db = ActorDatabase(user=src_user, password=src_password, database=src_db,
                                    db_host=src_host, logger=self.logger)

        # Initialize the destination database manager
        self.dest_db = DatabaseManager(user=dest_user, password=dest_password, database=dest_db,
                                       db_host=dest_host)

        self.batch_size = batch_size
        self.last_export_time = self.get_last_export_time()

    def get_last_export_time(self):
        """
        Reads the last export timestamp from a file.
        """
        if os.path.exists(LAST_EXPORT_FILE):
            with open(LAST_EXPORT_FILE, "r") as f:
                timestamp_str = f.read().strip()
                try:
                    return datetime.fromisoformat(timestamp_str)
                except ValueError:
                    self.logger.warning("Invalid timestamp in last export file, defaulting to epoch.")
        return datetime(1970, 1, 1, tzinfo=timezone.utc)  # Default to epoch

    def update_last_export_time(self, new_timestamp):
        """
        Updates the last export timestamp file with the current run time.
        """
        with open(LAST_EXPORT_FILE, "w") as f:
            f.write(new_timestamp.isoformat())

    def export(self):
        """
        Exports only the slices updated after the last execution timestamp.
        """
        try:
            self.logger.info(f"Starting export process... Last export was at {self.last_export_time}")

            offset = 0
            new_timestamp = datetime.now(timezone.utc)

            while True:
                self.logger.info(f"Fetching slices from offset {offset} (batch size: {self.batch_size})")
                slices = self.src_db.get_slices(offset=offset, limit=self.batch_size, slc_type=[SliceTypes.ClientSlice],
                                                updated_after=self.last_export_time)  # Fetch only updated slices

                if not slices or len(slices) == 0:
                    self.logger.info("No more slices to process. Export complete.")
                    break  # Stop when no more slices are found

                for slice_object in slices:
                    try:
                        project_id = self.dest_db.add_or_update_project(project_uuid=slice_object.get_project_id(),
                                                                        project_name=slice_object.get_project_name())
                        user_id = self.dest_db.add_or_update_user(user_uuid=slice_object.get_owner().get_oidc_sub_claim(),
                                                                  user_email=slice_object.get_owner().get_email())

                        slice_id = self.dest_db.add_or_update_slice(project_id=project_id, user_id=user_id,
                                                                    slice_guid=str(slice_object.get_slice_id()),
                                                                    slice_name=slice_object.get_name(),
                                                                    state=slice_object.get_state().value,
                                                                    lease_start=slice_object.get_lease_start(),
                                                                    lease_end=slice_object.get_lease_end())
                        for reservation in self.src_db.get_reservations(slice_id=slice_object.get_slice_id()):
                            error_message = reservation.get_error_message()
                            sliver = InventoryForType.get_allocated_sliver(reservation=reservation)
                            site_name = None
                            host_name = None
                            site_id = None
                            host_id = None
                            ip_subnet = None
                            core = None
                            ram = None
                            disk = None
                            image = None
                            bw = None

                            if isinstance(sliver, NodeSliver):
                                site_name = sliver.get_site()
                                if sliver.label_allocations and sliver.label_allocations.instance_parent:
                                    host_name = sliver.label_allocations.instance_parent
                                ip_subnet = str(sliver.management_ip)
                                image = sliver.image_ref

                                if sliver.capacity_allocations:
                                    core = sliver.capacity_allocations.core
                                    ram = sliver.capacity_allocations.ram
                                    disk = sliver.capacity_allocations.disk

                            elif isinstance(sliver, NetworkServiceSliver):
                                site_name = sliver.get_site()
                                if sliver.get_gateway():
                                    ip_subnet = str(sliver.get_gateway().subnet)
                                if sliver.capacities:
                                    bw = sliver.capacities.bw

                            if site_name:
                                site_id = self.dest_db.add_or_update_site(site_name=site_name.upper())
                                if host_name:
                                    host_id = self.dest_db.add_or_update_host(host_name=host_name.lower(), site_id=site_id)

                            sliver_id = self.dest_db.add_or_update_sliver(project_id=project_id,
                                                                          user_id=user_id,
                                                                          slice_id=slice_id,
                                                                          site_id=site_id,
                                                                          host_id=host_id,
                                                                          sliver_guid=str(reservation.get_reservation_id()),
                                                                          lease_start=reservation.get_term().get_start_time(),
                                                                          lease_end=reservation.get_term().get_end_time(),
                                                                          state=reservation.get_state().value,
                                                                          ip_subnet=ip_subnet,
                                                                          core=core,
                                                                          ram=ram,
                                                                          disk=disk,
                                                                          image=image,
                                                                          bandwidth=bw,
                                                                          sliver_type=str(reservation.get_type()).lower(),
                                                                          error=error_message)

                            if isinstance(sliver, NodeSliver) and sliver.attached_components_info:
                                for component in sliver.attached_components_info.devices.values():
                                    bdfs = component.labels.bdf if component.labels and component.labels.bdf else None
                                    if bdfs and not isinstance(bdfs, list):
                                        bdfs = [bdfs]
                                    self.dest_db.add_or_update_component(sliver_id=sliver_id,
                                                                         component_guid=component.node_id,
                                                                         component_type=str(component.get_type()).lower(),
                                                                         model=str(component.get_model()).lower(),
                                                                         bdfs=bdfs)

                            if isinstance(sliver, NetworkServiceSliver) and sliver.interface_info:
                                for ifs in sliver.interface_info.interfaces.values():
                                    vlan = ifs.labels.vlan if ifs.labels else None
                                    if not vlan and ifs.label_allocations:
                                        vlan = ifs.label_allocations.vlan

                                    bdf = ifs.labels.bdf if ifs.labels else None
                                    if not bdf and ifs.label_allocations:
                                        bdf = ifs.label_allocations.bdf

                                    local_name = ifs.labels.local_name if ifs.labels else None
                                    if not local_name and ifs.label_allocations:
                                        local_name = ifs.label_allocations.local_name

                                    device_name = ifs.labels.device_name if ifs.labels else None
                                    if not device_name and ifs.label_allocations:
                                        device_name = ifs.label_allocations.device_name

                                    self.dest_db.add_or_update_interface(sliver_id=sliver_id,
                                                                         interface_guid=ifs.node_id,
                                                                         vlan=vlan,
                                                                         name=ifs.get_name(),
                                                                         bdf=bdf,
                                                                         local_name=local_name,
                                                                         device_name=device_name)

                    except Exception as slice_error:
                        self.logger.error(f"Error processing slice {slice_object.get_slice_id()}: {slice_error}")
                        traceback.print_exc()
                offset += self.batch_size  # Move to the next batch

            self.logger.info(f"Updating last export time to {new_timestamp}")
            self.update_last_export_time(new_timestamp)
            self.logger.info("Export process completed successfully!")

        except Exception as e:
            self.logger.error(f"Exception occurred during export: {e}")
            traceback.print_exc()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Export data from Postgres to SQLAlchemy DB via DatabaseManager")
    parser.add_argument("--src_user", default="fabric", help="Source database username")
    parser.add_argument("--src_password", default="fabric", help="Source database password")
    parser.add_argument("--src_db", default="orchestrator", help="Source database name")
    parser.add_argument("--src_host", default="orchestrator-db:5432", help="Source database host")
    parser.add_argument("--dest_user", default="fabric", help="Destination database username")
    parser.add_argument("--dest_password", default="fabric", help="Destination database password")
    parser.add_argument("--dest_db", default="analytics", help="Destination database name")
    parser.add_argument("--dest_host", default="analytics-db:5432", help="Destination database host")
    parser.add_argument("--batch_size", type=int, default=1000, help="Number of slices to process per batch")

    args = parser.parse_args()

    exporter = ExportScript(
        src_user=args.src_user, src_password=args.src_password, src_db=args.src_db, src_host=args.src_host,
        dest_user=args.dest_user, dest_password=args.dest_password, dest_db=args.dest_db, dest_host=args.dest_host,
        batch_size=args.batch_size
    )

    exporter.export()



# coding: utf-8

from __future__ import absolute_import
from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from fabric_cf.orchestrator.swagger_server.models.base_model_ import Model
from fabric_cf.orchestrator.swagger_server import util


class SlicesPost(Model):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    def __init__(self, graph_model: str=None, ssh_keys: List[str]=None):  # noqa: E501
        """SlicesPost - a model defined in Swagger

        :param graph_model: The graph_model of this SlicesPost.  # noqa: E501
        :type graph_model: str
        :param ssh_keys: The ssh_keys of this SlicesPost.  # noqa: E501
        :type ssh_keys: List[str]
        """
        self.swagger_types = {
            'graph_model': str,
            'ssh_keys': List[str]
        }

        self.attribute_map = {
            'graph_model': 'graph_model',
            'ssh_keys': 'ssh_keys'
        }
        self._graph_model = graph_model
        self._ssh_keys = ssh_keys

    @classmethod
    def from_dict(cls, dikt) -> 'SlicesPost':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The slices_post of this SlicesPost.  # noqa: E501
        :rtype: SlicesPost
        """
        return util.deserialize_model(dikt, cls)

    @property
    def graph_model(self) -> str:
        """Gets the graph_model of this SlicesPost.


        :return: The graph_model of this SlicesPost.
        :rtype: str
        """
        return self._graph_model

    @graph_model.setter
    def graph_model(self, graph_model: str):
        """Sets the graph_model of this SlicesPost.


        :param graph_model: The graph_model of this SlicesPost.
        :type graph_model: str
        """
        if graph_model is None:
            raise ValueError("Invalid value for `graph_model`, must not be `None`")  # noqa: E501

        self._graph_model = graph_model

    @property
    def ssh_keys(self) -> List[str]:
        """Gets the ssh_keys of this SlicesPost.


        :return: The ssh_keys of this SlicesPost.
        :rtype: List[str]
        """
        return self._ssh_keys

    @ssh_keys.setter
    def ssh_keys(self, ssh_keys: List[str]):
        """Sets the ssh_keys of this SlicesPost.


        :param ssh_keys: The ssh_keys of this SlicesPost.
        :type ssh_keys: List[str]
        """
        if ssh_keys is None:
            raise ValueError("Invalid value for `ssh_keys`, must not be `None`")  # noqa: E501

        self._ssh_keys = ssh_keys



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import threading
import traceback
from typing import TYPE_CHECKING

from fabric_cf.actor.core.apis.abc_mgmt_client_actor import ABCMgmtClientActor
from fabric_cf.actor.core.apis.abc_mgmt_server_actor import ABCMgmtServerActor
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.core.actor_identity import ActorIdentity
from fabric_cf.actor.core.manage.messages.client_mng import ClientMng
from fabric_mb.message_bus.messages.proxy_avro import ProxyAvro
from fabric_cf.actor.core.util.id import ID
from fabric_cf.actor.core.apis.abc_actor_mixin import ActorType

if TYPE_CHECKING:
    from fabric_cf.actor.core.apis.abc_mgmt_actor import ABCMgmtActor
    from fabric_cf.actor.core.apis.abc_actor_mixin import ABCActorMixin


class RemoteActorCacheException(Exception):
    """
    Exception raised by RemoteActorCache
    """


class RemoteActorCache:
    """
    Maintains Remote Actors to which Actor is connected
    """
    actor_name = "NAME"
    actor_guid = "GUID"
    actor_location = "LOCATION"
    actor_protocol = "PROTOCOL"
    actor_type = "TYPE"

    def __init__(self):
        self.cache = {}
        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        self.logger = GlobalsSingleton.get().get_logger()
        self.lock = threading.Lock()
        self.local_actor_guids = set()

    def known_guids(self) -> set:
        """
        Returns list of know GUIDs
        """
        try:
            self.lock.acquire()
            result = set()
            for k in self.cache.keys():
                result.add(k)
            return result
        finally:
            self.lock.release()

    def check_to_remove_entry(self, *, guid: ID):
        """
        Check if actor entry can be removed
        @param guid actor guid
        """
        if guid is None:
            return

        try:
            self.lock.acquire()
            if guid in self.cache:
                value = self.cache[guid]
                if self.actor_location not in value:
                    self.cache.pop(guid)
        finally:
            self.lock.release()

    def add_cache_entry(self, *, guid: ID, entry: dict):
        """
        Add actor entry to Cache
        @param guid actor guid
        @param entry actor entry
        """
        if guid is None or entry is None:
            return

        try:
            self.lock.acquire()
            self.non_mt_cache_merge(guid=guid, entry=entry)
            self.local_actor_guids.add(guid)
        finally:
            self.lock.release()

    def add_partial_cache_entry(self, *, guid: ID, entry: dict):
        """
        Add partial actor entry to Cache
        @param guid actor guid
        @param entry actor entry
        """
        if guid is None or entry is None:
            return
        try:
            self.lock.acquire()
            self.non_mt_cache_merge(guid=guid, entry=entry)
        finally:
            self.lock.release()

    def non_mt_cache_merge(self, *, guid: ID, entry: dict):
        """
        Merge to an existing entry if present; remove otherwise
        @param guid actor guid
        @param entry actor entry
        """
        if guid not in self.cache:
            self.logger.debug("Inserting new entry for {}".format(guid))
            self.cache[guid] = entry
            return
        current = self.cache[guid]
        for k, v in entry.items():
            current[k] = v

        self.cache[guid] = current

    def get_cache_entry_copy(self, *, guid: ID) -> dict:
        """
        Get a copy of cacher entry
        @param guid actor guid
        """
        try:
            self.lock.acquire()
            if guid in self.cache:
                ret_val = self.cache[guid]
                return ret_val
        finally:
            self.lock.release()

    def check_peer(self, *, mgmt_actor: ABCMgmtActor, peer_guid: ID, peer_type: ActorType):
        """
        Check if a peer is already connected
        @param mgmt_actor mgmt_actor
        @param peer_guid peer_guid
        @param peer_type peer_type
        """
        self.logger.debug(f"Check if Peer {peer_guid}/{peer_type} already exists!")
        try:
            # For Broker - all AMs are added as proxies
            # For Orchestrator - all peers will be added as Proxies
            if isinstance(mgmt_actor, ABCMgmtClientActor) and peer_type in [ActorType.Authority, ActorType.Broker]:
                self.logger.debug(f"Checking brokers")
                brokers = mgmt_actor.get_brokers(broker=peer_guid)
                self.logger.debug(f"brokers -- {brokers}")
                if brokers is not None:
                    self.logger.debug(f"Edge between {mgmt_actor.get_guid()} and {peer_guid} exists (broker)")
                    return True

            # For AM - all peers will be added as clients
            # For Broker - orchestrator as client
            elif isinstance(mgmt_actor, ABCMgmtServerActor) and peer_type in [ActorType.Orchestrator, ActorType.Broker]:
                self.logger.debug(f"Checking clients")
                clients = mgmt_actor.get_clients(guid=peer_guid)
                self.logger.debug(f"clients -- {clients} {mgmt_actor.get_last_error()}")
                if clients is not None:
                    self.logger.debug(f"Edge between {mgmt_actor.get_guid()} and {peer_guid} exists (client)")
                    return True
        except Exception as e:
            raise RemoteActorCacheException(f"Unable to cast actor {mgmt_actor.get_guid()} or {peer_guid} e={e}")

        self.logger.debug(f"Edge between {mgmt_actor.get_guid()} and {peer_guid} does not exist")
        return False

    def establish_peer_private(self, *, mgmt_actor: ABCMgmtActor, peer_guid: ID, peer_type: ActorType,
                               update: bool = False) -> ClientMng:
        """
        Establish connection i.e. create either proxies or clients between peer
        @param mgmt_actor mgmt_actor
        @param peer_guid peer_guid
        @param peer_type peer_type
        @param update update
        """
        self.logger.debug("establish_peer_private IN")
        client = None
        cache_entry = self.get_cache_entry_copy(guid=peer_guid)
        if cache_entry is None:
            raise RemoteActorCacheException(f"Actor {peer_guid} does not have a registry cache entry")

        protocol = cache_entry.get(self.actor_protocol)
        kafka_topic = cache_entry.get(self.actor_location)
        identity = ActorIdentity(name=cache_entry.get(self.actor_name), guid=peer_guid)

        if kafka_topic is None:
            raise RemoteActorCacheException(f"Actor {peer_guid} does not have a kafka topic")

        if isinstance(mgmt_actor, ABCMgmtClientActor) and peer_type in [ActorType.Authority, ActorType.Broker]:
            proxy = ProxyAvro()
            proxy.set_protocol(protocol)
            proxy.set_guid(str(identity.get_guid()))
            proxy.set_name(identity.get_name())
            proxy.set_type(cache_entry.get(self.actor_type))
            proxy.set_kafka_topic(kafka_topic)

            try:
                if not update:
                    self.logger.debug(f"Creating new proxy: {proxy}")
                    if not mgmt_actor.add_broker(broker=proxy):
                        raise RemoteActorCacheException(f"Could not register broker {peer_guid} "
                                                        f"error: {mgmt_actor.get_last_error()}")
                else:
                    self.logger.debug(f"Updating existing proxy: {proxy}")
                    if not mgmt_actor.update_broker(broker=proxy):
                        raise RemoteActorCacheException(f"Could not update broker {peer_guid} "
                                                        f"error: {mgmt_actor.get_last_error()}")
            except Exception as e:
                self.logger.error(e)
                self.logger.error(traceback.format_exc())
        elif isinstance(mgmt_actor, ABCMgmtServerActor) and peer_type in [ActorType.Orchestrator, ActorType.Broker]:
            client = ClientMng()
            client.set_name(name=cache_entry.get(self.actor_name))
            client.set_guid(guid=str(peer_guid))
            try:
                if not update:
                    self.logger.debug(f"Creating new client: {client}")
                    mgmt_actor.register_client(client=client, kafka_topic=kafka_topic)
                else:
                    self.logger.debug(f"Updating existing client: {client}")
                    mgmt_actor.update_client(client=client, kafka_topic=kafka_topic)
            except Exception as e:
                raise RemoteActorCacheException(f"Could not register actor: {peer_guid} as a client of "
                                                f"actor: {mgmt_actor} e= {e}")

        self.logger.debug("establish_peer_private OUT {}".format(client))
        return client

    def establish_peer(self, *, mgmt_actor: ABCMgmtActor, peer_guid: ID, peer_type: ActorType) -> ClientMng:
        """
        Check if peer exists in cache and if not Establish connection i.e. create either proxies or clients between peer
        @param mgmt_actor mgmt_actor
        @param peer_guid peer_guid
        @param peer_type peer_type
        """
        self.logger.debug("establish_peer IN")
        client = None
        if mgmt_actor is None or peer_guid is None:
            self.logger.error("Cannot establish peer when either guid is not known")
            raise RemoteActorCacheException("Cannot establish peer when either guid is not known")
        try:
            update = self.check_peer(mgmt_actor=mgmt_actor, peer_guid=peer_guid, peer_type=peer_type)

            client = self.establish_peer_private(mgmt_actor=mgmt_actor, peer_guid=peer_guid, peer_type=peer_type,
                                                 update=update)

            self.check_to_remove_entry(guid=peer_guid)

            self.logger.debug(f"Peer established from {mgmt_actor} to {peer_guid}")

        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.logger.error(f"Peer could not be established from {mgmt_actor} to {peer_guid} e:={e}")
        self.logger.debug("establish_peer OUT {}".format(client))
        return client

    def register_with_registry(self, *, actor: ABCActorMixin):
        """
        Register an actor with Registry
        @param actor actor
        """
        try:
            act_name = actor.get_name()
            act_type = actor.get_type()
            act_guid = actor.get_guid()

            from fabric_cf.actor.core.container.globals import GlobalsSingleton
            kafka_topic = GlobalsSingleton.get().get_config().get_actor_config().get_kafka_topic()

            entry = {self.actor_name: act_name,
                     self.actor_guid: act_guid,
                     self.actor_type: act_type.name,
                     self.actor_protocol: Constants.PROTOCOL_KAFKA,
                     self.actor_location: kafka_topic}

            self.add_cache_entry(guid=act_guid, entry=entry)
            # TODO start liveness thread
        except Exception as e:
            self.logger.debug("Could not register actor {} with local registry e: {}".format(actor.get_name(), e))


class RemoteActorCacheSingleton:
    """
    Remote Actor Cache Singleton
    """
    __instance = None

    def __init__(self):
        if self.__instance is not None:
            raise RemoteActorCacheException("Singleton can't be created twice !")

    def get(self):
        """
        Actually create an instance
        """
        if self.__instance is None:
            self.__instance = RemoteActorCache()
        return self.__instance

    get = classmethod(get)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from datetime import timedelta, datetime, timezone
from http.client import BAD_REQUEST
from typing import List

from fabric_cf.actor.core.common.constants import Constants

from fabric_cf.orchestrator.core.exceptions import OrchestratorException
from fabric_cf.orchestrator.core.orchestrator_handler import OrchestratorHandler
from fabric_cf.orchestrator.swagger_server.models import Status200OkNoContentData, Slice, Sliver, SlicesPost
from fabric_cf.orchestrator.swagger_server.models.slice_details import SliceDetails  # noqa: E501
from fabric_cf.orchestrator.swagger_server.models.slices import Slices  # noqa: E501
from fabric_cf.orchestrator.swagger_server.models.slivers import Slivers  # noqa: E501
from fabric_cf.orchestrator.swagger_server.models.status200_ok_no_content import Status200OkNoContent  # noqa: E501
from fabric_cf.orchestrator.swagger_server import received_counter, success_counter, failure_counter
from fabric_cf.orchestrator.swagger_server.response.constants import POST_METHOD, SLICES_CREATE_PATH, DELETE_METHOD, \
    SLICES_DELETE_PATH, GET_METHOD, SLICES_GET_PATH, SLICES_RENEW_PATH, SLICES_GET_SLICE_ID_PATH, SLICES_MODIFY_PATH, \
    SLICES_MODIFY_ACCEPT_PATH, SLICES_DELETE_SLICE_ID_PATH
from fabric_cf.orchestrator.swagger_server.response.utils import get_token, cors_error_response, cors_success_response


def slices_creates_post(body: SlicesPost, name, lifetime=None, lease_start_time=None, lease_end_time=None):  # noqa: E501
    """Create slice

    Request to create a slice as described in the request, represented as a Graph ML specifying requested resources.
    Resources may be scheduled for immediate or future provisioning. On success, the allocated resources (slivers)
    are returned in Graph ML form and assigned to the requested slice. The orchestrator triggers asynchronous
    provisioning on appropriate sites based on the request timing. Experimenters can invoke the 'get slice' API to
    obtain the latest state of the requested resources.   # noqa: E501

    :param body: Contains the slice details, including requested resources.
    :type body: dict | bytes
    :param name: Slice Name
    :type name: str
    :param lifetime: Optional. The requested slice duration in hours.
    :type lifetime: int
    :param lease_start_time: Requested lease start time for the slice.
    :type lease_start_time: str
    :param lease_end_time: Requested lease end time for the slice.
    :type lease_end_time: str

    :rtype: Slivers
    """
    handler = OrchestratorHandler()
    logger = handler.get_logger()
    received_counter.labels(POST_METHOD, SLICES_CREATE_PATH).inc()

    try:
        token = get_token()
        ssh_key = ','.join(body.ssh_keys)
        start = handler.validate_lease_time(lease_time=lease_start_time)
        end = handler.validate_lease_time(lease_time=lease_end_time)
        now = datetime.now(timezone.utc)

        # Check that start time is at least 60 minutes in the future
        if start and (start - now) < timedelta(minutes=60):
            raise OrchestratorException(
                http_error_code=BAD_REQUEST,
                message="Requested Start Time should be at least 60 minutes from the current time!"
            )

        # Check for valid lease duration between start and end times
        if start and end:
            diff = end - start
            if diff < timedelta(minutes=60):
                raise OrchestratorException(
                    http_error_code=BAD_REQUEST,
                    message="The requested lease duration must be at least 60 minutes."
                )

            max_duration_hours = Constants.DEFAULT_MAX_DURATION_IN_WEEKS.total_seconds() / 3600  # Convert weeks to hours
            if diff > timedelta(hours=max_duration_hours):
                raise OrchestratorException(
                    http_error_code=BAD_REQUEST,
                    message=f"Requested lease duration exceeds the maximum allowed duration of "
                            f"{max_duration_hours} hours."
                )

        # Set the lifetime to 24 hours if not specified and calculate based on start and end times
        if start and end:
            hours = int((end - start).total_seconds() / 3600)
            if not lifetime:
                lifetime = Constants.DEFAULT_LEASE_IN_HOURS  # Default to 24 hours if unspecified

            # Ensure lifetime does not exceed the specified lease duration
            if lifetime > hours:
                raise OrchestratorException(
                    http_error_code=BAD_REQUEST,
                    message="The specified lifetime exceeds the allowable duration between the start and end times."
                )

        # Create the slice and assemble the response
        slivers_dict = handler.create_slice(token=token, slice_name=name, slice_graph=body.graph_model,
                                            lease_start_time=start, lease_end_time=end,
                                            ssh_key=ssh_key, lifetime=lifetime)
        response = Slivers()
        response.data = [Sliver().from_dict(s) for s in slivers_dict]
        response.size = len(response.data)
        response.type = "slivers"
        success_counter.labels(POST_METHOD, SLICES_CREATE_PATH).inc()
        return cors_success_response(response_body=response)

    except OrchestratorException as e:
        logger.exception(e)
        failure_counter.labels(POST_METHOD, SLICES_CREATE_PATH).inc()
        return cors_error_response(error=e)
    except Exception as e:
        logger.exception(e)
        failure_counter.labels(POST_METHOD, SLICES_CREATE_PATH).inc()
        return cors_error_response(error=e)


def slices_delete_delete():  # noqa: E501
    """Delete all slices for a User within a project.

    Delete all slices for a User within a project. User identity email and project id is available in the bearer token.  # noqa: E501


    :rtype: Status200OkNoContent
    """
    handler = OrchestratorHandler()
    logger = handler.get_logger()
    received_counter.labels(DELETE_METHOD, SLICES_DELETE_PATH).inc()
    try:
        token = get_token()
        handler.delete_slices(token=token)
        success_counter.labels(DELETE_METHOD, SLICES_DELETE_PATH).inc()

        slice_info = Status200OkNoContentData()
        slice_info.details = f"Slices for user have been successfully deleted"
        response = Status200OkNoContent()
        response.data = [slice_info]
        response.size = len(response.data)
        response.status = 200
        response.type = 'no_content'
        return cors_success_response(response_body=response)

    except OrchestratorException as e:
        logger.exception(e)
        failure_counter.labels(DELETE_METHOD, SLICES_DELETE_PATH).inc()
        return cors_error_response(error=e)
    except Exception as e:
        logger.exception(e)
        failure_counter.labels(DELETE_METHOD, SLICES_DELETE_PATH).inc()
        return cors_error_response(error=e)


def slices_delete_slice_id_delete(slice_id) -> Status200OkNoContent:  # noqa: E501
    """Delete slice.

    Request to delete slice. On success, resources associated with slice or sliver are stopped if necessary,
    de-provisioned and un-allocated at the respective sites.  # noqa: E501

    :param slice_id: Slice identified by universally unique identifier
    :type slice_id: str

    :rtype: Status200OkNoContent
    """
    handler = OrchestratorHandler()
    logger = handler.get_logger()
    received_counter.labels(DELETE_METHOD, SLICES_DELETE_SLICE_ID_PATH).inc()
    try:
        token = get_token()
        handler.delete_slices(token=token, slice_id=slice_id)
        success_counter.labels(DELETE_METHOD, SLICES_DELETE_SLICE_ID_PATH).inc()

        slice_info = Status200OkNoContentData()
        slice_info.details = f"Slice '{slice_id}' has been successfully deleted"
        response = Status200OkNoContent()
        response.data = [slice_info]
        response.size = len(response.data)
        response.status = 200
        response.type = 'no_content'
        return cors_success_response(response_body=response)

    except OrchestratorException as e:
        logger.exception(e)
        failure_counter.labels(DELETE_METHOD, SLICES_DELETE_SLICE_ID_PATH).inc()
        return cors_error_response(error=e)
    except Exception as e:
        logger.exception(e)
        failure_counter.labels(DELETE_METHOD, SLICES_DELETE_SLICE_ID_PATH).inc()
        return cors_error_response(error=e)


def slices_get(name: str = None, search: str = None, exact_match: bool = False,
               as_self: bool = True, states: List[str] = None, limit: int = 5, offset: int = 0):  # noqa: E501
    """Retrieve a listing of user slices

    Retrieve a listing of user slices. It returns list of all slices belonging to all members in a project when
    &#x27;as_self&#x27; is False otherwise returns only the all user&#x27;s slices in a project. # noqa: E501

    :param name: Search for Slices with the name
    :type name: str
    :param search: search term applied
    :type search: str
    :param exact_match: Exact Match for Search term
    :type exact_match: str
    :param as_self: GET object as Self
    :type as_self: bool
    :param states: Search for Slices in the specified states
    :type states: List[str]
    :param limit: maximum number of results to return per page (1 or more)
    :type limit: int
    :param offset: number of items to skip before starting to collect the result set
    :type offset: int

    :rtype: Slices
    """
    handler = OrchestratorHandler()
    logger = handler.get_logger()
    received_counter.labels(GET_METHOD, SLICES_GET_PATH).inc()
    try:
        token = get_token()
        slices_dict = handler.get_slices(token=token, states=states, name=name, limit=limit, offset=offset,
                                         as_self=as_self, search=search, exact_match=exact_match)
        response = Slices()
        response.data = []
        response.type = 'slices'
        for s in slices_dict:
            slice_obj = Slice().from_dict(s)
            response.data.append(slice_obj)
        response.size = len(response.data)

        success_counter.labels(GET_METHOD, SLICES_GET_PATH).inc()
        return cors_success_response(response_body=response)
    except OrchestratorException as e:
        logger.exception(e)
        failure_counter.labels(GET_METHOD, SLICES_GET_PATH).inc()
        return cors_error_response(error=e)
    except Exception as e:
        logger.exception(e)
        failure_counter.labels(GET_METHOD, SLICES_GET_PATH).inc()
        return cors_error_response(error=e)


def slices_modify_slice_id_accept_post(slice_id):  # noqa: E501
    """Accept the last modify an existing slice

    Accept the last modify and prune any failed resources from the Slice.
    Also return the accepted slice model back to the user.   # noqa: E501

    :param slice_id: Slice identified by universally unique identifier
    :type slice_id: str

    :rtype: Slivers
    """
    handler = OrchestratorHandler()
    logger = handler.get_logger()
    received_counter.labels(POST_METHOD, SLICES_MODIFY_ACCEPT_PATH).inc()
    try:
        token = get_token()
        value = handler.modify_accept(token=token, slice_id=slice_id)
        slice_object = Slice().from_dict(value)
        response = SliceDetails(data=[slice_object], size=1)
        response.type = 'slice_details'
        success_counter.labels(POST_METHOD, SLICES_MODIFY_ACCEPT_PATH).inc()
        return cors_success_response(response_body=response)
    except OrchestratorException as e:
        logger.exception(e)
        failure_counter.labels(POST_METHOD, SLICES_MODIFY_ACCEPT_PATH).inc()
        return cors_error_response(error=e)
    except Exception as e:
        logger.exception(e)
        failure_counter.labels(POST_METHOD, SLICES_MODIFY_ACCEPT_PATH).inc()
        return cors_error_response(error=e)


def slices_modify_slice_id_put(body, slice_id):  # noqa: E501
    """Modify an existing slice

    Request to modify an existing slice as described in the request. Request would be a graph ML describing the
    experiment topolgy expected after a modify. The supported modify actions include adding or removing nodes,
    components, network services or interfaces of the slice. On success, one or more slivers are allocated,
    containing resources satisfying the request, and assigned to the given slice. This API returns list and
    description of the resources reserved for the slice in the form of Graph ML. Orchestrator would also trigger
    provisioning of these resources asynchronously on the appropriate sites either now or in the future as requested.
    Experimenter can invoke get slice API to get the latest state of the requested resources.   # noqa: E501

    :param body:
    :type body: dict | bytes
    :param slice_id: Slice identified by universally unique identifier
    :type slice_id: str

    :rtype: Slivers
    """
    handler = OrchestratorHandler()
    logger = handler.get_logger()
    received_counter.labels(POST_METHOD, SLICES_MODIFY_PATH).inc()

    try:
        token = get_token()
        slice_graph = body.decode("utf-8")
        slivers_dict = handler.modify_slice(token=token, slice_id=slice_id, slice_graph=slice_graph)
        response = Slivers()
        response.data = []
        for s in slivers_dict:
            sliver = Sliver().from_dict(s)
            response.data.append(sliver)
        response.size = len(response.data)
        response.type = "slivers"
        success_counter.labels(POST_METHOD, SLICES_MODIFY_PATH).inc()
        return cors_success_response(response_body=response)
    except OrchestratorException as e:
        logger.exception(e)
        failure_counter.labels(POST_METHOD, SLICES_MODIFY_PATH).inc()
        return cors_error_response(error=e)
    except Exception as e:
        logger.exception(e)
        failure_counter.labels(POST_METHOD, SLICES_CREATE_PATH).inc()
        return cors_error_response(error=e)


def slices_renew_slice_id_post(slice_id, lease_end_time) -> Status200OkNoContent:  # noqa: E501
    """Renew slice

    Request to extend slice be renewed with their expiration extended. If possible, the orchestrator should extend the
    slivers to the requested expiration time, or to a sooner time if policy limits apply.  # noqa: E501

    :param slice_id: Slice identified by universally unique identifier
    :type slice_id: str
    :param lease_end_time: New Lease End Time for the Slice
    :type lease_end_time: str

    :rtype: Status200OkNoContent
    """
    handler = OrchestratorHandler()
    logger = handler.get_logger()
    received_counter.labels(POST_METHOD, SLICES_RENEW_PATH).inc()

    try:
        token = get_token()
        end = handler.validate_lease_time(lease_time=lease_end_time)
        handler.renew_slice(token=token, slice_id=slice_id, new_lease_end_time=end)
        success_counter.labels(POST_METHOD, SLICES_RENEW_PATH).inc()

        slice_info = Status200OkNoContentData()
        slice_info.details = f"Slice '{slice_id}' has been successfully renewed"
        response = Status200OkNoContent()
        response.data = [slice_info]
        response.size = len(response.data)
        response.status = 200
        response.type = 'no_content'
        return cors_success_response(response_body=response)
    except OrchestratorException as e:
        logger.exception(e)
        failure_counter.labels(POST_METHOD, SLICES_RENEW_PATH).inc()
        return cors_error_response(error=e)
    except Exception as e:
        logger.exception(e)
        failure_counter.labels(POST_METHOD, SLICES_RENEW_PATH).inc()
        return cors_error_response(error=e)


def slices_slice_id_get(slice_id, graph_format, as_self=True) -> SliceDetails:  # noqa: E501
    """slice properties

    Retrieve Slice properties # noqa: E501

    :param slice_id: Slice identified by universally unique identifier
    :type slice_id: str
    :param graph_format: graph format
    :type graph_format: str
    :param as_self: GET object as Self
    :type as_self: bool

    :rtype: SliceDetails
    """
    handler = OrchestratorHandler()
    logger = handler.get_logger()
    received_counter.labels(GET_METHOD, SLICES_GET_SLICE_ID_PATH).inc()
    try:
        token = get_token()
        value = handler.get_slice_graph(token=token, slice_id=slice_id, graph_format_str=graph_format,
                                        as_self=as_self)
        slice_object = Slice().from_dict(value)
        response = SliceDetails(data=[slice_object], size=1)
        response.type = 'slice_details'
        success_counter.labels(GET_METHOD, SLICES_GET_SLICE_ID_PATH).inc()
        return cors_success_response(response_body=response)
    except OrchestratorException as e:
        logger.exception(e)
        failure_counter.labels(GET_METHOD, SLICES_GET_SLICE_ID_PATH).inc()
        return cors_error_response(error=e)
    except Exception as e:
        logger.exception(e)
        failure_counter.labels(GET_METHOD, SLICES_GET_SLICE_ID_PATH).inc()
        return cors_error_response(error=e)
